# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the generic buyer skill."""

from pathlib import Path
from typing import cast

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.connections.ledger.base import CONNECTION_ID as LEDGER_PUBLIC_ID
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_buyer.behaviours import GenericSearchBehaviour
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy

from tests.conftest import ROOT_DIR


FETCHAI = "fetchai"


class TestSkillBehaviour(BaseSkillTestCase):
    """Test behaviours of generic buyer."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "generic_buyer")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.search_behaviour = cast(
            GenericSearchBehaviour, cls._skill.skill_context.behaviours.search
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)

    def test_setup_is_ledger_tx(self):
        """Test the setup method of the search behaviour where is_ledger_tx is True."""
        # operation
        self.search_behaviour.setup()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            to=str(LEDGER_PUBLIC_ID),
            sender=self.skill.skill_context.agent_address,
            ledger_id=FETCHAI,
            address=self.skill.skill_context.agent_address,
        )
        assert has_attributes, error_str

    def test_setup_not_is_ledger_tx(self):
        """Test the setup method of the search behaviour where is_ledger_tx is False."""
        # setup
        self.strategy._is_ledger_tx = False

        # before
        assert not self.strategy.is_searching

        # operation
        self.search_behaviour.setup()

        # after
        assert self.strategy.is_searching

    def test_act_is_searching(self):
        """Test the act method of the search behaviour where is_searching is True."""
        # setup
        self.strategy._is_searching = True

        # operation
        self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            to=self.skill.skill_context.search_service_address,
            sender=self.skill.skill_context.agent_address,
            query=self.skill.skill_context.strategy.get_location_and_service_query(),
        )
        assert has_attributes, error_str

    def test_act_not_is_searching(self):
        """Test the act method of the search behaviour where is_searching is False."""
        # setup
        self.strategy._is_searching = False

        # operation
        self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)

    def test_teardown(self):
        """Test the teardown method of the search behaviour."""
        assert self.search_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains a generic data model."""

from typing import Any, Dict, List

from aea.exceptions import enforce
from aea.helpers.search.models import Attribute, DataModel, Location


SUPPORTED_TYPES = {"str": str, "int": int, "float": float, "bool": bool}


class GenericDataModel(DataModel):  # pylint: disable=too-few-public-methods
    """Generic data model."""

    def __init__(self, data_model_name: str, data_model_attributes: Dict[str, Any]):
        """Initialise the dataModel."""
        self.attributes = []  # type: List[Attribute]
        for values in data_model_attributes.values():
            enforce(
                values["type"] in SUPPORTED_TYPES.keys(),
                "Type is not supported. Use str, int, float or bool",
            )
            enforce(isinstance(values["name"], str), "Name must be a string!")
            enforce(
                isinstance(values["is_required"], bool),
                "Wrong type for is_required. Must be bool!",
            )
            self.attributes.append(
                Attribute(
                    name=values["name"],  # type: ignore
                    type_=SUPPORTED_TYPES[values["type"]],
                    is_required=values["is_required"],
                )
            )

        super().__init__(data_model_name, self.attributes)


AGENT_LOCATION_MODEL = DataModel(
    "location_agent",
    [Attribute("location", Location, True, "The location where the agent is.")],
    "A data model to describe location of an agent.",
)


AGENT_PERSONALITY_MODEL = DataModel(
    "personality_agent",
    [
        Attribute("piece", str, True, "The personality piece key."),
        Attribute("value", str, True, "The personality piece value."),
    ],
    "A data model to describe the personality of an agent.",
)


AGENT_SET_SERVICE_MODEL = DataModel(
    "set_service_key",
    [
        Attribute("key", str, True, "Service key name."),
        Attribute("value", str, True, "Service key value."),
    ],
    "A data model to set service key.",
)


SIMPLE_SERVICE_MODEL = DataModel(
    "simple_service",
    [Attribute("seller_service", str, True, "Service key name.")],
    "A data model to represent a search for a service.",
)


SIMPLE_DATA_MODEL = DataModel(
    "simple_data",
    [Attribute("dataset_id", str, True, "Data set key name.")],
    "A data model to represent a search for a data set.",
)


AGENT_REMOVE_SERVICE_MODEL = DataModel(
    "remove_service_key",
    [Attribute("key", str, True, "Service key name.")],
    "A data model to remove service key.",
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains multiaddress class."""

from binascii import unhexlify
from typing import Optional

import base58
import multihash  # type: ignore
from ecdsa import VerifyingKey, curves, keys

from aea.helpers.multiaddr.crypto_pb2 import KeyType, PublicKey


# NOTE:
# - Reference: https://github.com/libp2p/specs/blob/master/peer-ids/peer-ids.md#keys
# - Implementation inspired from https://github.com/libp2p/py-libp2p
# - On inlining see: https://github.com/libp2p/specs/issues/138
# - Enabling inlining to be interoperable w/ the Go implementation

ENABLE_INLINING = True
MAX_INLINE_KEY_LENGTH = 42
IDENTITY_MULTIHASH_CODE = 0x00

KEY_SIZE = 32
ZERO = b"\x00"

if ENABLE_INLINING:

    class IdentityHash:
        """Neutral hashing implementation for inline multihashing."""

        _digest: bytes

        def __init__(self) -> None:
            """Initialize IdentityHash object."""
            self._digest = bytearray()

        def update(self, input_data: bytes) -> None:
            """
            Update data to hash.

            :param input_data: the data
            :return: None
            """
            self._digest += input_data

        def digest(self) -> bytes:
            """
            Get hash of input data.

            :return: the hash
            """
            return self._digest

    multihash.FuncReg.register(
        IDENTITY_MULTIHASH_CODE, "identity", hash_new=IdentityHash
    )


def _pad_scalar(scalar):
    """Pad scalar."""
    return (ZERO * (KEY_SIZE - len(scalar))) + scalar


def _pad_hex(hexed):
    """Pad odd-length hex strings."""
    return hexed if not len(hexed) & 1 else "0" + hexed


def _hex_to_bytes(hexed):
    """Hex to bytes."""
    return _pad_scalar(unhexlify(_pad_hex(hexed)))


class MultiAddr:
    """Protocol Labs' Multiaddress representation of a network address."""

    def __init__(
        self,
        host: str,
        port: int,
        public_key: Optional[str] = None,
        multihash_id: Optional[str] = None,
    ):
        """
        Initialize a multiaddress.

        :param host: ip host of the address
        :param port: port number of the address
        :param public_key: hex encoded public key. Must conform to Bitcoin EC encoding standard for Secp256k1
        :param multihash_id: a multihash of the public key
        """

        self._host = host
        self._port = port

        if public_key is not None:
            try:
                VerifyingKey._from_compressed(
                    _hex_to_bytes(public_key), curves.SECP256k1
                )
            except keys.MalformedPointError as e:  # pragma: no cover
                raise ValueError(
                    "Malformed public key '{}': {}".format(public_key, str(e))
                )

            self._public_key = public_key
            self._peerid = self.compute_peerid(self._public_key)
        elif multihash_id is not None:
            try:
                multihash.decode(base58.b58decode(multihash_id))
            except Exception as e:
                raise ValueError(
                    "Malformed multihash '{}': {}".format(multihash_id, str(e))
                )

            self._public_key = ""
            self._peerid = multihash_id
        else:
            raise ValueError(  # pragma: no cover
                "MultiAddr requires either public_key or multihash_id to be provided."
            )

    @staticmethod
    def compute_peerid(public_key: str) -> str:
        """
        Compute the peer id from a public key.

        In particular, compute the base58 representation of
        libp2p PeerID from Bitcoin EC encoded Secp256k1 public key.

        :param public_key: the public key.
        :return: the peer id.
        """
        key_protobuf = PublicKey(
            key_type=KeyType.Secp256k1, data=_hex_to_bytes(public_key)  # type: ignore
        )
        key_serialized = key_protobuf.SerializeToString()
        algo = multihash.Func.sha2_256
        if ENABLE_INLINING and len(key_serialized) <= MAX_INLINE_KEY_LENGTH:
            algo = IDENTITY_MULTIHASH_CODE
        key_mh = multihash.digest(key_serialized, algo)
        return base58.b58encode(key_mh.encode()).decode()

    @classmethod
    def from_string(cls, maddr: str) -> "MultiAddr":
        """
        Construct a MultiAddr object from its string format

        :param maddr: multiaddress string
        """
        parts = maddr.split("/")
        if len(parts) != 7 or not parts[4].isdigit():
            raise ValueError("Malformed multiaddress '{}'".format(maddr))

        return cls(host=parts[2], port=int(parts[4]), multihash_id=parts[6])

    @property
    def public_key(self) -> str:
        """Get the public key."""
        return self._public_key

    @property
    def peer_id(self) -> str:
        """Get the peer id."""
        return self._peerid

    @property
    def host(self) -> str:
        """Get the peer host."""
        return self._host

    @property
    def port(self) -> int:
        """Get the peer port."""
        return self._port

    def format(self) -> str:
        """Canonical representation of a multiaddress."""
        return f"/dns4/{self._host}/tcp/{self._port}/p2p/{self._peerid}"

    def __str__(self) -> str:
        """Default string representation of a mutliaddress."""
        return self.format()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tac's message definition."""

import logging
from typing import Dict, Optional, Set, Tuple, cast

from aea.configurations.base import ProtocolId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message

from packages.fetchai.protocols.tac.custom_types import ErrorCode as CustomErrorCode


_default_logger = logging.getLogger("aea.packages.fetchai.protocols.tac.message")

DEFAULT_BODY_SIZE = 4


class TacMessage(Message):
    """The tac protocol implements the messages an AEA needs to participate in the TAC."""

    protocol_id = ProtocolId.from_str("fetchai/tac:0.8.0")

    ErrorCode = CustomErrorCode

    class Performative(Message.Performative):
        """Performatives for the tac protocol."""

        CANCELLED = "cancelled"
        GAME_DATA = "game_data"
        REGISTER = "register"
        TAC_ERROR = "tac_error"
        TRANSACTION = "transaction"
        TRANSACTION_CONFIRMATION = "transaction_confirmation"
        UNREGISTER = "unregister"

        def __str__(self):
            """Get the string representation."""
            return str(self.value)

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs,
    ):
        """
        Initialise an instance of TacMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        """
        self._performatives = {
            "cancelled",
            "game_data",
            "register",
            "tac_error",
            "transaction",
            "transaction_confirmation",
            "unregister",
        }
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=TacMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(TacMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def agent_addr_to_name(self) -> Dict[str, str]:
        """Get the 'agent_addr_to_name' content from the message."""
        enforce(
            self.is_set("agent_addr_to_name"),
            "'agent_addr_to_name' content is not set.",
        )
        return cast(Dict[str, str], self.get("agent_addr_to_name"))

    @property
    def agent_name(self) -> str:
        """Get the 'agent_name' content from the message."""
        enforce(self.is_set("agent_name"), "'agent_name' content is not set.")
        return cast(str, self.get("agent_name"))

    @property
    def amount_by_currency_id(self) -> Dict[str, int]:
        """Get the 'amount_by_currency_id' content from the message."""
        enforce(
            self.is_set("amount_by_currency_id"),
            "'amount_by_currency_id' content is not set.",
        )
        return cast(Dict[str, int], self.get("amount_by_currency_id"))

    @property
    def counterparty_address(self) -> str:
        """Get the 'counterparty_address' content from the message."""
        enforce(
            self.is_set("counterparty_address"),
            "'counterparty_address' content is not set.",
        )
        return cast(str, self.get("counterparty_address"))

    @property
    def counterparty_signature(self) -> str:
        """Get the 'counterparty_signature' content from the message."""
        enforce(
            self.is_set("counterparty_signature"),
            "'counterparty_signature' content is not set.",
        )
        return cast(str, self.get("counterparty_signature"))

    @property
    def currency_id_to_name(self) -> Dict[str, str]:
        """Get the 'currency_id_to_name' content from the message."""
        enforce(
            self.is_set("currency_id_to_name"),
            "'currency_id_to_name' content is not set.",
        )
        return cast(Dict[str, str], self.get("currency_id_to_name"))

    @property
    def error_code(self) -> CustomErrorCode:
        """Get the 'error_code' content from the message."""
        enforce(self.is_set("error_code"), "'error_code' content is not set.")
        return cast(CustomErrorCode, self.get("error_code"))

    @property
    def exchange_params_by_currency_id(self) -> Dict[str, float]:
        """Get the 'exchange_params_by_currency_id' content from the message."""
        enforce(
            self.is_set("exchange_params_by_currency_id"),
            "'exchange_params_by_currency_id' content is not set.",
        )
        return cast(Dict[str, float], self.get("exchange_params_by_currency_id"))

    @property
    def fee_by_currency_id(self) -> Dict[str, int]:
        """Get the 'fee_by_currency_id' content from the message."""
        enforce(
            self.is_set("fee_by_currency_id"),
            "'fee_by_currency_id' content is not set.",
        )
        return cast(Dict[str, int], self.get("fee_by_currency_id"))

    @property
    def good_id_to_name(self) -> Dict[str, str]:
        """Get the 'good_id_to_name' content from the message."""
        enforce(self.is_set("good_id_to_name"), "'good_id_to_name' content is not set.")
        return cast(Dict[str, str], self.get("good_id_to_name"))

    @property
    def info(self) -> Optional[Dict[str, str]]:
        """Get the 'info' content from the message."""
        return cast(Optional[Dict[str, str]], self.get("info"))

    @property
    def ledger_id(self) -> str:
        """Get the 'ledger_id' content from the message."""
        enforce(self.is_set("ledger_id"), "'ledger_id' content is not set.")
        return cast(str, self.get("ledger_id"))

    @property
    def nonce(self) -> str:
        """Get the 'nonce' content from the message."""
        enforce(self.is_set("nonce"), "'nonce' content is not set.")
        return cast(str, self.get("nonce"))

    @property
    def quantities_by_good_id(self) -> Dict[str, int]:
        """Get the 'quantities_by_good_id' content from the message."""
        enforce(
            self.is_set("quantities_by_good_id"),
            "'quantities_by_good_id' content is not set.",
        )
        return cast(Dict[str, int], self.get("quantities_by_good_id"))

    @property
    def sender_address(self) -> str:
        """Get the 'sender_address' content from the message."""
        enforce(self.is_set("sender_address"), "'sender_address' content is not set.")
        return cast(str, self.get("sender_address"))

    @property
    def sender_signature(self) -> str:
        """Get the 'sender_signature' content from the message."""
        enforce(
            self.is_set("sender_signature"), "'sender_signature' content is not set."
        )
        return cast(str, self.get("sender_signature"))

    @property
    def transaction_id(self) -> str:
        """Get the 'transaction_id' content from the message."""
        enforce(self.is_set("transaction_id"), "'transaction_id' content is not set.")
        return cast(str, self.get("transaction_id"))

    @property
    def utility_params_by_good_id(self) -> Dict[str, float]:
        """Get the 'utility_params_by_good_id' content from the message."""
        enforce(
            self.is_set("utility_params_by_good_id"),
            "'utility_params_by_good_id' content is not set.",
        )
        return cast(Dict[str, float], self.get("utility_params_by_good_id"))

    @property
    def version_id(self) -> str:
        """Get the 'version_id' content from the message."""
        enforce(self.is_set("version_id"), "'version_id' content is not set.")
        return cast(str, self.get("version_id"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the tac protocol."""
        try:
            enforce(
                type(self.dialogue_reference) == tuple,
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                type(self.dialogue_reference[0]) == str,
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                type(self.dialogue_reference[1]) == str,
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) == int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) == int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                type(self.performative) == TacMessage.Performative,
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self.body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == TacMessage.Performative.REGISTER:
                expected_nb_of_contents = 1
                enforce(
                    type(self.agent_name) == str,
                    "Invalid type for content 'agent_name'. Expected 'str'. Found '{}'.".format(
                        type(self.agent_name)
                    ),
                )
            elif self.performative == TacMessage.Performative.UNREGISTER:
                expected_nb_of_contents = 0
            elif self.performative == TacMessage.Performative.TRANSACTION:
                expected_nb_of_contents = 10
                enforce(
                    type(self.transaction_id) == str,
                    "Invalid type for content 'transaction_id'. Expected 'str'. Found '{}'.".format(
                        type(self.transaction_id)
                    ),
                )
                enforce(
                    type(self.ledger_id) == str,
                    "Invalid type for content 'ledger_id'. Expected 'str'. Found '{}'.".format(
                        type(self.ledger_id)
                    ),
                )
                enforce(
                    type(self.sender_address) == str,
                    "Invalid type for content 'sender_address'. Expected 'str'. Found '{}'.".format(
                        type(self.sender_address)
                    ),
                )
                enforce(
                    type(self.counterparty_address) == str,
                    "Invalid type for content 'counterparty_address'. Expected 'str'. Found '{}'.".format(
                        type(self.counterparty_address)
                    ),
                )
                enforce(
                    type(self.amount_by_currency_id) == dict,
                    "Invalid type for content 'amount_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.amount_by_currency_id)
                    ),
                )
                for (
                    key_of_amount_by_currency_id,
                    value_of_amount_by_currency_id,
                ) in self.amount_by_currency_id.items():
                    enforce(
                        type(key_of_amount_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'amount_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_amount_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_amount_by_currency_id) == int,
                        "Invalid type for dictionary values in content 'amount_by_currency_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_amount_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.fee_by_currency_id) == dict,
                    "Invalid type for content 'fee_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.fee_by_currency_id)
                    ),
                )
                for (
                    key_of_fee_by_currency_id,
                    value_of_fee_by_currency_id,
                ) in self.fee_by_currency_id.items():
                    enforce(
                        type(key_of_fee_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'fee_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_fee_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_fee_by_currency_id) == int,
                        "Invalid type for dictionary values in content 'fee_by_currency_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_fee_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.quantities_by_good_id) == dict,
                    "Invalid type for content 'quantities_by_good_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.quantities_by_good_id)
                    ),
                )
                for (
                    key_of_quantities_by_good_id,
                    value_of_quantities_by_good_id,
                ) in self.quantities_by_good_id.items():
                    enforce(
                        type(key_of_quantities_by_good_id) == str,
                        "Invalid type for dictionary keys in content 'quantities_by_good_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_quantities_by_good_id)
                        ),
                    )
                    enforce(
                        type(value_of_quantities_by_good_id) == int,
                        "Invalid type for dictionary values in content 'quantities_by_good_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_quantities_by_good_id)
                        ),
                    )
                enforce(
                    type(self.nonce) == str,
                    "Invalid type for content 'nonce'. Expected 'str'. Found '{}'.".format(
                        type(self.nonce)
                    ),
                )
                enforce(
                    type(self.sender_signature) == str,
                    "Invalid type for content 'sender_signature'. Expected 'str'. Found '{}'.".format(
                        type(self.sender_signature)
                    ),
                )
                enforce(
                    type(self.counterparty_signature) == str,
                    "Invalid type for content 'counterparty_signature'. Expected 'str'. Found '{}'.".format(
                        type(self.counterparty_signature)
                    ),
                )
            elif self.performative == TacMessage.Performative.CANCELLED:
                expected_nb_of_contents = 0
            elif self.performative == TacMessage.Performative.GAME_DATA:
                expected_nb_of_contents = 9
                enforce(
                    type(self.amount_by_currency_id) == dict,
                    "Invalid type for content 'amount_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.amount_by_currency_id)
                    ),
                )
                for (
                    key_of_amount_by_currency_id,
                    value_of_amount_by_currency_id,
                ) in self.amount_by_currency_id.items():
                    enforce(
                        type(key_of_amount_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'amount_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_amount_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_amount_by_currency_id) == int,
                        "Invalid type for dictionary values in content 'amount_by_currency_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_amount_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.exchange_params_by_currency_id) == dict,
                    "Invalid type for content 'exchange_params_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.exchange_params_by_currency_id)
                    ),
                )
                for (
                    key_of_exchange_params_by_currency_id,
                    value_of_exchange_params_by_currency_id,
                ) in self.exchange_params_by_currency_id.items():
                    enforce(
                        type(key_of_exchange_params_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'exchange_params_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_exchange_params_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_exchange_params_by_currency_id) == float,
                        "Invalid type for dictionary values in content 'exchange_params_by_currency_id'. Expected 'float'. Found '{}'.".format(
                            type(value_of_exchange_params_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.quantities_by_good_id) == dict,
                    "Invalid type for content 'quantities_by_good_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.quantities_by_good_id)
                    ),
                )
                for (
                    key_of_quantities_by_good_id,
                    value_of_quantities_by_good_id,
                ) in self.quantities_by_good_id.items():
                    enforce(
                        type(key_of_quantities_by_good_id) == str,
                        "Invalid type for dictionary keys in content 'quantities_by_good_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_quantities_by_good_id)
                        ),
                    )
                    enforce(
                        type(value_of_quantities_by_good_id) == int,
                        "Invalid type for dictionary values in content 'quantities_by_good_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_quantities_by_good_id)
                        ),
                    )
                enforce(
                    type(self.utility_params_by_good_id) == dict,
                    "Invalid type for content 'utility_params_by_good_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.utility_params_by_good_id)
                    ),
                )
                for (
                    key_of_utility_params_by_good_id,
                    value_of_utility_params_by_good_id,
                ) in self.utility_params_by_good_id.items():
                    enforce(
                        type(key_of_utility_params_by_good_id) == str,
                        "Invalid type for dictionary keys in content 'utility_params_by_good_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_utility_params_by_good_id)
                        ),
                    )
                    enforce(
                        type(value_of_utility_params_by_good_id) == float,
                        "Invalid type for dictionary values in content 'utility_params_by_good_id'. Expected 'float'. Found '{}'.".format(
                            type(value_of_utility_params_by_good_id)
                        ),
                    )
                enforce(
                    type(self.fee_by_currency_id) == dict,
                    "Invalid type for content 'fee_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.fee_by_currency_id)
                    ),
                )
                for (
                    key_of_fee_by_currency_id,
                    value_of_fee_by_currency_id,
                ) in self.fee_by_currency_id.items():
                    enforce(
                        type(key_of_fee_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'fee_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_fee_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_fee_by_currency_id) == int,
                        "Invalid type for dictionary values in content 'fee_by_currency_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_fee_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.agent_addr_to_name) == dict,
                    "Invalid type for content 'agent_addr_to_name'. Expected 'dict'. Found '{}'.".format(
                        type(self.agent_addr_to_name)
                    ),
                )
                for (
                    key_of_agent_addr_to_name,
                    value_of_agent_addr_to_name,
                ) in self.agent_addr_to_name.items():
                    enforce(
                        type(key_of_agent_addr_to_name) == str,
                        "Invalid type for dictionary keys in content 'agent_addr_to_name'. Expected 'str'. Found '{}'.".format(
                            type(key_of_agent_addr_to_name)
                        ),
                    )
                    enforce(
                        type(value_of_agent_addr_to_name) == str,
                        "Invalid type for dictionary values in content 'agent_addr_to_name'. Expected 'str'. Found '{}'.".format(
                            type(value_of_agent_addr_to_name)
                        ),
                    )
                enforce(
                    type(self.currency_id_to_name) == dict,
                    "Invalid type for content 'currency_id_to_name'. Expected 'dict'. Found '{}'.".format(
                        type(self.currency_id_to_name)
                    ),
                )
                for (
                    key_of_currency_id_to_name,
                    value_of_currency_id_to_name,
                ) in self.currency_id_to_name.items():
                    enforce(
                        type(key_of_currency_id_to_name) == str,
                        "Invalid type for dictionary keys in content 'currency_id_to_name'. Expected 'str'. Found '{}'.".format(
                            type(key_of_currency_id_to_name)
                        ),
                    )
                    enforce(
                        type(value_of_currency_id_to_name) == str,
                        "Invalid type for dictionary values in content 'currency_id_to_name'. Expected 'str'. Found '{}'.".format(
                            type(value_of_currency_id_to_name)
                        ),
                    )
                enforce(
                    type(self.good_id_to_name) == dict,
                    "Invalid type for content 'good_id_to_name'. Expected 'dict'. Found '{}'.".format(
                        type(self.good_id_to_name)
                    ),
                )
                for (
                    key_of_good_id_to_name,
                    value_of_good_id_to_name,
                ) in self.good_id_to_name.items():
                    enforce(
                        type(key_of_good_id_to_name) == str,
                        "Invalid type for dictionary keys in content 'good_id_to_name'. Expected 'str'. Found '{}'.".format(
                            type(key_of_good_id_to_name)
                        ),
                    )
                    enforce(
                        type(value_of_good_id_to_name) == str,
                        "Invalid type for dictionary values in content 'good_id_to_name'. Expected 'str'. Found '{}'.".format(
                            type(value_of_good_id_to_name)
                        ),
                    )
                enforce(
                    type(self.version_id) == str,
                    "Invalid type for content 'version_id'. Expected 'str'. Found '{}'.".format(
                        type(self.version_id)
                    ),
                )
                if self.is_set("info"):
                    expected_nb_of_contents += 1
                    info = cast(Dict[str, str], self.info)
                    enforce(
                        type(info) == dict,
                        "Invalid type for content 'info'. Expected 'dict'. Found '{}'.".format(
                            type(info)
                        ),
                    )
                    for key_of_info, value_of_info in info.items():
                        enforce(
                            type(key_of_info) == str,
                            "Invalid type for dictionary keys in content 'info'. Expected 'str'. Found '{}'.".format(
                                type(key_of_info)
                            ),
                        )
                        enforce(
                            type(value_of_info) == str,
                            "Invalid type for dictionary values in content 'info'. Expected 'str'. Found '{}'.".format(
                                type(value_of_info)
                            ),
                        )
            elif self.performative == TacMessage.Performative.TRANSACTION_CONFIRMATION:
                expected_nb_of_contents = 3
                enforce(
                    type(self.transaction_id) == str,
                    "Invalid type for content 'transaction_id'. Expected 'str'. Found '{}'.".format(
                        type(self.transaction_id)
                    ),
                )
                enforce(
                    type(self.amount_by_currency_id) == dict,
                    "Invalid type for content 'amount_by_currency_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.amount_by_currency_id)
                    ),
                )
                for (
                    key_of_amount_by_currency_id,
                    value_of_amount_by_currency_id,
                ) in self.amount_by_currency_id.items():
                    enforce(
                        type(key_of_amount_by_currency_id) == str,
                        "Invalid type for dictionary keys in content 'amount_by_currency_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_amount_by_currency_id)
                        ),
                    )
                    enforce(
                        type(value_of_amount_by_currency_id) == int,
                        "Invalid type for dictionary values in content 'amount_by_currency_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_amount_by_currency_id)
                        ),
                    )
                enforce(
                    type(self.quantities_by_good_id) == dict,
                    "Invalid type for content 'quantities_by_good_id'. Expected 'dict'. Found '{}'.".format(
                        type(self.quantities_by_good_id)
                    ),
                )
                for (
                    key_of_quantities_by_good_id,
                    value_of_quantities_by_good_id,
                ) in self.quantities_by_good_id.items():
                    enforce(
                        type(key_of_quantities_by_good_id) == str,
                        "Invalid type for dictionary keys in content 'quantities_by_good_id'. Expected 'str'. Found '{}'.".format(
                            type(key_of_quantities_by_good_id)
                        ),
                    )
                    enforce(
                        type(value_of_quantities_by_good_id) == int,
                        "Invalid type for dictionary values in content 'quantities_by_good_id'. Expected 'int'. Found '{}'.".format(
                            type(value_of_quantities_by_good_id)
                        ),
                    )
            elif self.performative == TacMessage.Performative.TAC_ERROR:
                expected_nb_of_contents = 1
                enforce(
                    type(self.error_code) == CustomErrorCode,
                    "Invalid type for content 'error_code'. Expected 'ErrorCode'. Found '{}'.".format(
                        type(self.error_code)
                    ),
                )
                if self.is_set("info"):
                    expected_nb_of_contents += 1
                    info = cast(Dict[str, str], self.info)
                    enforce(
                        type(info) == dict,
                        "Invalid type for content 'info'. Expected 'dict'. Found '{}'.".format(
                            type(info)
                        ),
                    )
                    for key_of_info, value_of_info in info.items():
                        enforce(
                            type(key_of_info) == str,
                            "Invalid type for dictionary keys in content 'info'. Expected 'str'. Found '{}'.".format(
                                type(key_of_info)
                            ),
                        )
                        enforce(
                            type(value_of_info) == str,
                            "Invalid type for dictionary values in content 'info'. Expected 'str'. Found '{}'.".format(
                                type(value_of_info)
                            ),
                        )

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
            else:
                enforce(
                    0 < self.target < self.message_id,
                    "Invalid 'target'. Expected an integer between 1 and {} inclusive. Found {}.".format(
                        self.message_id - 1, self.target,
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea add-key` sub-command."""
import os
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase, mock

import yaml

import aea
from aea.cli import cli
from aea.cli.add_key import _try_add_key
from aea.configurations.base import AgentConfig, DEFAULT_AEA_CONFIG_FILE

from tests.conftest import (
    AUTHOR,
    CLI_LOG_OPTION,
    CUR_PATH,
    CliRunner,
    ETHEREUM,
    ETHEREUM_PRIVATE_KEY_FILE,
    FETCHAI,
    FETCHAI_PRIVATE_KEY_FILE,
    ROOT_DIR,
)
from tests.test_cli.tools_for_testing import ContextMock


class TestAddFetchKey:
    """Test that the command 'aea add-key' works as expected for a 'fetchai' key."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.agent_folder = Path(cls.t, cls.agent_name)
        os.chdir(cls.t)

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "create", "--local", cls.agent_name]
        )
        assert result.exit_code == 0
        os.chdir(Path(cls.t, cls.agent_name))

        shutil.copy(
            Path(CUR_PATH, "data", FETCHAI_PRIVATE_KEY_FILE),
            cls.agent_folder / FETCHAI_PRIVATE_KEY_FILE,
        )

        cls.result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "add-key", FETCHAI, FETCHAI_PRIVATE_KEY_FILE],
        )

    def test_return_code(self):
        """Test return code equal to zero."""
        assert self.result.exit_code == 0

    def test_key_added(self):
        """Test that the fetch private key has been added correctly."""
        f = open(Path(self.agent_folder, DEFAULT_AEA_CONFIG_FILE))
        expected_json = yaml.safe_load(f)
        config = AgentConfig.from_json(expected_json)
        private_key_path = config.private_key_paths.read(FETCHAI)
        assert private_key_path == FETCHAI_PRIVATE_KEY_FILE
        assert len(config.private_key_paths.read_all()) == 1

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        shutil.rmtree(cls.t)


class TestAddEthereumhKey:
    """Test that the command 'aea add-key' works as expected for an 'ethereum' key."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.agent_folder = Path(cls.t, cls.agent_name)
        os.chdir(cls.t)

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "create", "--local", cls.agent_name]
        )
        assert result.exit_code == 0
        os.chdir(Path(cls.t, cls.agent_name))

        shutil.copy(
            Path(CUR_PATH, "data", ETHEREUM_PRIVATE_KEY_FILE),
            cls.agent_folder / ETHEREUM_PRIVATE_KEY_FILE,
        )

        cls.result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "add-key", ETHEREUM, ETHEREUM_PRIVATE_KEY_FILE],
        )

    def test_return_code(self):
        """Test return code equal to zero."""
        assert self.result.exit_code == 0

    def test_key_added(self):
        """Test that the fetch private key has been added correctly."""
        f = open(Path(self.agent_folder, DEFAULT_AEA_CONFIG_FILE))
        expected_json = yaml.safe_load(f)
        config = AgentConfig.from_json(expected_json)
        private_key_path = config.private_key_paths.read(ETHEREUM)
        assert private_key_path == ETHEREUM_PRIVATE_KEY_FILE
        assert len(config.private_key_paths.read_all()) == 1

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        shutil.rmtree(cls.t)


class TestAddManyKeys:
    """Test that the command 'aea add-key' works as expected when adding many keys."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.agent_folder = Path(cls.t, cls.agent_name)
        os.chdir(cls.t)

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )

        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "create", "--local", cls.agent_name]
        )
        assert result.exit_code == 0
        os.chdir(Path(cls.t, cls.agent_name))

        shutil.copy(
            Path(CUR_PATH, "data", FETCHAI_PRIVATE_KEY_FILE),
            cls.agent_folder / FETCHAI_PRIVATE_KEY_FILE,
        )
        shutil.copy(
            Path(CUR_PATH, "data", ETHEREUM_PRIVATE_KEY_FILE),
            cls.agent_folder / ETHEREUM_PRIVATE_KEY_FILE,
        )

    def test_add_many_keys(self, pytestconfig):
        """Test that the keys are added correctly."""

        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "add-key", FETCHAI, FETCHAI_PRIVATE_KEY_FILE],
        )
        assert result.exit_code == 0
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "add-key", ETHEREUM, ETHEREUM_PRIVATE_KEY_FILE],
        )
        assert result.exit_code == 0

        f = open(Path(self.agent_folder, DEFAULT_AEA_CONFIG_FILE))
        expected_json = yaml.safe_load(f)
        config = AgentConfig.from_json(expected_json)
        private_key_path_ethereum = config.private_key_paths.read(FETCHAI)
        assert private_key_path_ethereum == FETCHAI_PRIVATE_KEY_FILE
        private_key_path_ethereum = config.private_key_paths.read(ETHEREUM)
        assert private_key_path_ethereum == ETHEREUM_PRIVATE_KEY_FILE
        assert len(config.private_key_paths.read_all()) == 2

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except OSError:
            pass


def test_add_key_fails_bad_key():
    """Test that 'aea add-key' fails because the key is not valid."""
    oldcwd = os.getcwd()
    runner = CliRunner()
    agent_name = "myagent"
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)
    try:
        with mock.patch.object(
            aea.crypto.helpers._default_logger, "error"
        ) as mock_logger_error:

            result = runner.invoke(
                cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
            )

            result = runner.invoke(
                cli, [*CLI_LOG_OPTION, "create", "--local", agent_name]
            )
            assert result.exit_code == 0
            os.chdir(Path(tmpdir, agent_name))

            # create an empty file - surely not a private key
            pvk_file = "this_is_not_a_key.txt"
            Path(pvk_file).touch()

            result = runner.invoke(cli, [*CLI_LOG_OPTION, "add-key", FETCHAI, pvk_file])
            assert result.exit_code == 1
            error_message = "Invalid length of private key, received 0, expected 32"
            mock_logger_error.assert_called_with(
                "This is not a valid private key file: '{}'\n Exception: '{}'".format(
                    pvk_file, error_message
                ),
                exc_info=True,
            )

            # check that no key has been added.
            f = open(Path(DEFAULT_AEA_CONFIG_FILE))
            expected_json = yaml.safe_load(f)
            config = AgentConfig.from_json(expected_json)
            assert len(config.private_key_paths.read_all()) == 0
    finally:
        os.chdir(oldcwd)


def test_add_key_fails_bad_ledger_id():
    """Test that 'aea add-key' fails because the ledger id is not valid."""
    oldcwd = os.getcwd()
    runner = CliRunner()
    agent_name = "myagent"
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)
    try:
        result = runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )

        result = runner.invoke(cli, [*CLI_LOG_OPTION, "create", "--local", agent_name])
        assert result.exit_code == 0
        os.chdir(Path(tmpdir, agent_name))

        # generate a private key file
        result = runner.invoke(cli, [*CLI_LOG_OPTION, "generate-key", FETCHAI])
        assert result.exit_code == 0
        assert Path(FETCHAI_PRIVATE_KEY_FILE).exists()
        bad_ledger_id = "this_is_a_bad_ledger_id"

        result = runner.invoke(
            cli, [*CLI_LOG_OPTION, "add-key", bad_ledger_id, FETCHAI_PRIVATE_KEY_FILE]
        )
        assert result.exit_code == 2

        # check that no key has been added.
        f = open(Path(DEFAULT_AEA_CONFIG_FILE))
        expected_json = yaml.safe_load(f)
        config = AgentConfig.from_json(expected_json)
        assert len(config.private_key_paths.read_all()) == 0
    finally:
        os.chdir(oldcwd)


@mock.patch("builtins.open", mock.mock_open())
class AddKeyTestCase(TestCase):
    """Test case for _add_key method."""

    def test__add_key_positive(self, *mocks):
        """Test for _add_key method positive result."""
        ctx = ContextMock()
        _try_add_key(ctx, "type", "filepath")


@mock.patch("builtins.open", mock.mock_open())
class AddKeyConnectionTestCase(TestCase):
    """Test case for _add_key method."""

    def test__add_key_positive(self, *mocks):
        """Test for _add_key method positive result."""
        ctx = ContextMock()
        _try_add_key(ctx, "type", "filepath", connection=True)


@mock.patch("aea.cli.utils.decorators.try_to_load_agent_config")
@mock.patch("aea.cli.add_key.try_validate_private_key_path")
@mock.patch("aea.cli.add_key._try_add_key")
class AddKeyCommandTestCase(TestCase):
    """Test case for CLI add_key command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_run_positive(self, *mocks):
        """Test for CLI add_key positive result."""
        filepath = str(
            Path(ROOT_DIR, "setup.py")
        )  # some existing filepath to pass CLI argument check
        result = self.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "--skip-consistency-check",
                "add-key",
                FETCHAI,
                filepath,
            ],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of AEA agents project configuiration."""
import os
from shutil import rmtree
from typing import Dict, List, Set

from aea.aea import AEA
from aea.aea_builder import AEABuilder
from aea.cli.registry.fetch import fetch_agent
from aea.cli.utils.context import Context
from aea.configurations.base import PublicId


class Project:
    """Agent project representation."""

    def __init__(self, public_id: PublicId, path: str):
        """Init project with public_id and project's path."""
        self.public_id: PublicId = public_id
        self.path: str = path
        self.agents: Set[str] = set()

    @classmethod
    def load(cls, working_dir: str, public_id: PublicId) -> "Project":
        """Load project with given public_id to working_dir."""
        ctx = Context(cwd=working_dir)
        path = os.path.join(working_dir, public_id.author, public_id.name)
        fetch_agent(
            ctx, public_id, target_dir=os.path.join(public_id.author, public_id.name)
        )
        return cls(public_id, path)

    def remove(self) -> None:
        """Remove project, do cleanup."""
        rmtree(self.path)


class AgentAlias:
    """Agent alias representation."""

    def __init__(
        self,
        project: Project,
        agent_name: str,
        config: List[Dict],
        agent: AEA,
        builder: AEABuilder,
    ):
        """Init agent alias with project, config, name, agent, builder."""
        self.project = project
        self.config = config
        self.agent_name = agent_name
        self.agent = agent
        self.builder = builder
        self.project.agents.add(self.agent_name)

    def remove_from_project(self):
        """Remove agent alias from project."""
        self.project.agents.remove(self.agent_name)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea launch` sub-command."""

import logging
import os
import shutil
import sys
import tempfile
import unittest
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, List, Optional

import pytest
import yaml
from pexpect.exceptions import EOF  # type: ignore

from aea.cli import cli
from aea.configurations.base import DEFAULT_AEA_CONFIG_FILE

from tests.common.pexpect_popen import PexpectWrapper
from tests.conftest import AUTHOR, CLI_LOG_OPTION, CUR_PATH, CliRunner, MAX_FLAKY_RERUNS


logger = logging.getLogger(__name__)


class BaseLaunchTestCase:
    """Base Test case for launch tests."""

    @contextmanager
    def _cli_launch(
        self, agents: List[str], options: Optional[List[str]] = None
    ) -> Generator:
        """
        Run aea.cli wrapped with Pexpect.

        :param agents: list of agent names to run
        :param options: list of string options to pass to aea launch.

        :return: PexpectWrapper
        """
        proc = PexpectWrapper(  # nosec
            [
                sys.executable,
                "-m",
                "aea.cli",
                "-v",
                "DEBUG",
                "launch",
                *(options or []),
                *(agents or []),
            ],
            env=os.environ,
            maxread=10000,
            encoding="utf-8",
            logfile=sys.stdout,
        )
        try:
            yield proc
        finally:
            proc.wait_to_complete(10)

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        if cls is BaseLaunchTestCase:
            raise unittest.SkipTest("Skip BaseTest tests, it's a base class")

        cls.runner = CliRunner()
        cls.agent_name_1 = "myagent_1"
        cls.agent_name_2 = "myagent_2"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "create", "--local", cls.agent_name_1]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "create", "--local", cls.agent_name_2]
        )
        assert result.exit_code == 0

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestLaunch(BaseLaunchTestCase):
    """Test that the command 'aea launch <agent_name>' works as expected."""

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        with self._cli_launch([self.agent_name_1, self.agent_name_2]) as process_launch:
            process_launch.expect_all(
                [
                    f"[{self.agent_name_1}] Start processing messages...",
                    f"[{self.agent_name_2}] Start processing messages...",
                ],
                timeout=20,
            )
            process_launch.control_c()
            process_launch.expect_all(
                ["Exit cli. code: 0"], timeout=20,
            )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
class TestLaunchWithOneFailingAgent(BaseLaunchTestCase):
    """Test aea launch when there is a failing agent.."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        super().setup_class()
        # add the exception skill to agent 2
        os.chdir(cls.agent_name_2)
        shutil.copytree(
            Path(CUR_PATH, "data", "exception_skill"),
            Path(cls.t, cls.agent_name_2, "skills", "exception"),
        )
        config_path = Path(cls.t, cls.agent_name_2, DEFAULT_AEA_CONFIG_FILE)
        config = yaml.safe_load(open(config_path))
        config.setdefault("skills", []).append("fetchai/exception:0.1.0")
        yaml.safe_dump(config, open(config_path, "w"))
        os.chdir(cls.t)

    def test_exit_code_equal_to_one(self):
        """Assert that the exit code is equal to one (i.e. generic failure)."""
        with self._cli_launch([self.agent_name_1, self.agent_name_2]) as process_launch:
            process_launch.expect_all(
                [
                    f"[{self.agent_name_1}] Start processing messages...",
                    "Expected exception!",
                    "Receiving loop terminated",  # cause race condition in close/interrupt agent 2, so wait it closed by exception before call ctrl+c
                ],
                timeout=20,
            )
            process_launch.control_c()
            process_launch.expect_all(
                [
                    f"Agent {self.agent_name_1} terminated with exit code 0",
                    f"Agent {self.agent_name_2} terminated with exit code ",
                ],
                timeout=20,
            )
            process_launch.expect(
                EOF, timeout=20,
            )
            process_launch.wait_to_complete(10)
            assert process_launch.returncode == 1


class TestLaunchWithWrongArguments(BaseLaunchTestCase):
    """Test aea launch when some agent directory does not exist."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "launch", "this_agent_does_not_exist"],
            standalone_mode=True,
        )

    def test_exit_code_equal_to_one(self):
        """Assert that the exit code is equal to 1."""
        assert self.result.exit_code == 1


class TestLaunchMultithreaded(BaseLaunchTestCase):
    """Test that the command 'aea launch <agent_names> --multithreaded' works as expected."""

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        with self._cli_launch(
            [self.agent_name_1, self.agent_name_2], ["--multithreaded"]
        ) as process_launch:
            process_launch.expect_all(
                [
                    f"[{self.agent_name_1}] Start processing messages",
                    f"[{self.agent_name_2}] Start processing messages",
                ],
                timeout=30,
            )
            process_launch.control_c()
            process_launch.expect_all(
                ["Exit cli. code: 0"], timeout=30,
            )


class TestLaunchOneAgent(BaseLaunchTestCase):
    """Test that the command 'aea launch <agent_name>' works as expected."""

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        with self._cli_launch([self.agent_name_1]) as process_launch:
            process_launch.expect_all(
                [f"[{self.agent_name_1}] Start processing messages..."], timeout=20
            )
            process_launch.control_c()
            process_launch.expect_all(
                [f"Agent {self.agent_name_1} terminated with exit code 0"], timeout=20
            )
            process_launch.wait_to_complete(10)

            assert process_launch.returncode == 0



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains generic tools for AEA end-to-end testing."""

from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, cast

from aea.cli.utils.config import handle_dotted_path
from aea.configurations.base import (
    CRUDCollection,
    ComponentConfiguration,
    PackageConfiguration,
    PackageType,
    PublicId,
    SkillConfig,
    dependencies_from_json,
)
from aea.connections.stub.connection import write_envelope
from aea.exceptions import enforce
from aea.helpers.yaml_utils import yaml_dump, yaml_dump_all
from aea.mail.base import Envelope
from aea.test_tools.constants import DEFAULT_AUTHOR


def write_envelope_to_file(envelope: Envelope, file_path: str) -> None:
    """
    Write an envelope to a file.

    :param envelope: Envelope.
    :param file_path: the file path

    :return: None
    """
    with open(Path(file_path), "ab+") as f:
        write_envelope(envelope, f)


def read_envelope_from_file(file_path: str):
    """
    Read an envelope from a file.

    :param file_path the file path.

    :return: envelope
    """
    lines = []
    with open(Path(file_path), "rb+") as f:
        lines.extend(f.readlines())

    enforce(len(lines) == 2, "Did not find two lines.")
    line = lines[0] + lines[1]
    to_b, sender_b, protocol_id_b, message, end = line.strip().split(b",", maxsplit=4)
    to = to_b.decode("utf-8")
    sender = sender_b.decode("utf-8")
    protocol_id = PublicId.from_str(protocol_id_b.decode("utf-8"))
    enforce(end in [b"", b"\n"], "Envelope improperly formatted.")

    return Envelope(to=to, sender=sender, protocol_id=protocol_id, message=message,)


def _nested_set(
    configuration_obj: PackageConfiguration, keys: List, value: Any
) -> None:
    """
    Nested set a value to a dict. Force sets the values, overwriting any present values, but maintaining schema validation.

    :param configuration_obj: configuration object
    :param keys: list of keys.
    :param value: a value to set.

    :return: None.
    """

    def get_nested_ordered_dict_from_dict(input_dict: Dict) -> Dict:
        _dic = {}
        for _key, _value in input_dict.items():
            if isinstance(_value, dict):
                _dic[_key] = OrderedDict(get_nested_ordered_dict_from_dict(_value))
            else:
                _dic[_key] = _value
        return _dic

    def get_nested_ordered_dict_from_keys_and_value(
        keys: List[str], value: Any
    ) -> Dict:
        _dic = (
            OrderedDict(get_nested_ordered_dict_from_dict(value))
            if isinstance(value, dict)
            else value
        )
        for key in keys[::-1]:
            _dic = OrderedDict({key: _dic})
        return _dic

    root_key = keys[0]
    if (
        isinstance(configuration_obj, SkillConfig)
        and root_key in SkillConfig.FIELDS_WITH_NESTED_FIELDS
    ):
        root_attr = getattr(configuration_obj, root_key)
        length = len(keys)
        if length < 3 or keys[2] not in SkillConfig.NESTED_FIELDS_ALLOWED_TO_UPDATE:
            raise ValueError(f"Invalid keys={keys}.")  # pragma: nocover
        skill_component_id = keys[1]
        skill_component_config = root_attr.read(skill_component_id)
        if length == 3 and isinstance(value, dict):  # root.skill_component_id.args
            # set all args
            skill_component_config.args = get_nested_ordered_dict_from_dict(value)
        elif len(keys) >= 4:  # root.skill_component_id.args.[keys]
            # update some args
            dic = get_nested_ordered_dict_from_keys_and_value(keys[3:], value)
            skill_component_config.args.update(dic)
        else:
            raise ValueError(  # pragma: nocover
                f"Invalid keys={keys} and values={value}."
            )
        root_attr.update(skill_component_id, skill_component_config)
    else:
        root_attr = getattr(configuration_obj, root_key)
        if isinstance(root_attr, CRUDCollection):
            if isinstance(value, dict) and len(keys) == 1:  # root.
                for _key, _value in value.items():
                    dic = get_nested_ordered_dict_from_keys_and_value([_key], _value)
                    root_attr.update(_key, dic[_key])
            elif len(keys) >= 2:  # root.[keys]
                dic = get_nested_ordered_dict_from_keys_and_value(keys[1:], value)
                root_attr.update(keys[1], dic[keys[1]])
            else:
                raise ValueError(  # pragma: nocover
                    f"Invalid keys={keys} and values={value}."
                )
        elif root_key == "dependencies":
            enforce(
                isinstance(configuration_obj, ComponentConfiguration),
                "Cannot only set dependencies to ComponentConfiguration instances.",
            )
            configuration_obj = cast(ComponentConfiguration, configuration_obj)
            new_pypi_dependencies = dependencies_from_json(value)
            configuration_obj.pypi_dependencies = new_pypi_dependencies
        else:
            dic = get_nested_ordered_dict_from_keys_and_value(keys, value)
            setattr(configuration_obj, root_key, dic[root_key])


def nested_set_config(
    dotted_path: str, value: Any, author: str = DEFAULT_AUTHOR
) -> None:
    """
    Set an AEA config with nested values.

    Run from agent's directory.

    Allowed dotted_path:
        'agent.an_attribute_name'
        'protocols.my_protocol.an_attribute_name'
        'connections.my_connection.an_attribute_name'
        'contracts.my_contract.an_attribute_name'
        'skills.my_skill.an_attribute_name'
        'vendor.author.[protocols|connections|skills].package_name.attribute_name

    :param dotted_path: dotted path to a setting.
    :param value: a value to assign. Must be of yaml serializable type.
    :param author: the author name, used to parse the dotted path.

    :return: None.
    """
    settings_keys, config_file_path, config_loader, _ = handle_dotted_path(
        dotted_path, author
    )

    with config_file_path.open() as fp:
        config = config_loader.load(fp)

    _nested_set(config, settings_keys, value)

    if config.package_type == PackageType.AGENT:
        json_data = config.ordered_json
        component_configurations = json_data.pop("component_configurations")
        yaml_dump_all(
            [json_data] + component_configurations, config_file_path.open("w")
        )
    else:
        yaml_dump(config.ordered_json, config_file_path.open("w"))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the strategy class."""

import time
from typing import Dict

from temper import Temper

from packages.fetchai.skills.generic_seller.strategy import GenericStrategy


MAX_RETRIES = 10


class Strategy(GenericStrategy):
    """This class defines a strategy for the agent."""

    def collect_from_data_source(self) -> Dict[str, str]:
        """
        Build the data payload.

        :return: the data
        """
        temper = Temper()
        retries = 0
        while retries < MAX_RETRIES:
            results = temper.read()
            if "internal temperature" in results[0].keys():
                degrees = {"thermometer_data": str(results)}
                break
            self.context.logger.debug("Couldn't read the sensor I am re-trying.")
            time.sleep(0.5)
            retries += 1
        return degrees



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the integration test for the generic buyer and seller skills."""

from random import uniform

import pytest

from aea.test_tools.test_cases import AEATestCaseMany

from packages.fetchai.connections.p2p_libp2p.connection import LIBP2P_SUCCESS_MESSAGE

from tests.conftest import (
    COSMOS,
    COSMOS_PRIVATE_KEY_FILE_CONNECTION,
    FETCHAI,
    FETCHAI_PRIVATE_KEY_FILE,
    MAX_FLAKY_RERUNS_INTEGRATION,
    NON_FUNDED_COSMOS_PRIVATE_KEY_1,
    NON_GENESIS_CONFIG,
    wait_for_localhost_ports_to_close,
)


@pytest.mark.integration
class TestGenericSkills(AEATestCaseMany):
    """Test that generic skills work."""

    @pytest.mark.flaky(
        reruns=MAX_FLAKY_RERUNS_INTEGRATION
    )  # cause possible network issues
    def test_generic(self, pytestconfig):
        """Run the generic skills sequence."""
        seller_aea_name = "my_generic_seller"
        buyer_aea_name = "my_generic_buyer"
        self.create_agents(seller_aea_name, buyer_aea_name)

        default_routing = {
            "fetchai/ledger_api:0.5.0": "fetchai/ledger:0.7.0",
            "fetchai/oef_search:0.8.0": "fetchai/soef:0.10.0",
        }

        # generate random location
        location = {
            "latitude": round(uniform(-90, 90), 2),  # nosec
            "longitude": round(uniform(-180, 180), 2),  # nosec
        }

        # prepare seller agent
        self.set_agent_context(seller_aea_name)
        self.add_item("connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/soef:0.10.0")
        self.set_config("agent.default_connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/ledger:0.7.0")
        self.add_item("skill", "fetchai/generic_seller:0.14.0")
        setting_path = (
            "vendor.fetchai.skills.generic_seller.models.strategy.args.is_ledger_tx"
        )
        self.set_config(setting_path, False, "bool")
        setting_path = "agent.default_routing"
        self.nested_set_config(setting_path, default_routing)
        self.run_install()

        # add keys
        self.generate_private_key(FETCHAI)
        self.generate_private_key(COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION)
        self.add_private_key(FETCHAI, FETCHAI_PRIVATE_KEY_FILE)
        self.add_private_key(
            COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION, connection=True
        )
        self.replace_private_key_in_file(
            NON_FUNDED_COSMOS_PRIVATE_KEY_1, COSMOS_PRIVATE_KEY_FILE_CONNECTION
        )

        setting_path = "vendor.fetchai.connections.p2p_libp2p.config.ledger_id"
        self.set_config(setting_path, COSMOS)

        # make runable:
        setting_path = "vendor.fetchai.skills.generic_seller.is_abstract"
        self.set_config(setting_path, False, "bool")

        # replace location
        setting_path = (
            "vendor.fetchai.skills.generic_seller.models.strategy.args.location"
        )
        self.nested_set_config(setting_path, location)

        # prepare buyer agent
        self.set_agent_context(buyer_aea_name)
        self.add_item("connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/soef:0.10.0")
        self.set_config("agent.default_connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/ledger:0.7.0")
        self.add_item("skill", "fetchai/generic_buyer:0.13.0")
        setting_path = (
            "vendor.fetchai.skills.generic_buyer.models.strategy.args.is_ledger_tx"
        )
        self.set_config(setting_path, False, "bool")
        setting_path = "agent.default_routing"
        self.nested_set_config(setting_path, default_routing)
        self.run_install()

        # add keys
        self.generate_private_key(FETCHAI)
        self.generate_private_key(COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION)
        self.add_private_key(FETCHAI, FETCHAI_PRIVATE_KEY_FILE)
        self.add_private_key(
            COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION, connection=True
        )

        # set p2p configs
        setting_path = "vendor.fetchai.connections.p2p_libp2p.config"
        self.nested_set_config(setting_path, NON_GENESIS_CONFIG)

        # make runable:
        setting_path = "vendor.fetchai.skills.generic_buyer.is_abstract"
        self.set_config(setting_path, False, "bool")

        # replace location
        setting_path = (
            "vendor.fetchai.skills.generic_buyer.models.strategy.args.location"
        )
        self.nested_set_config(setting_path, location)

        # run AEAs
        self.set_agent_context(seller_aea_name)
        seller_aea_process = self.run_agent()

        check_strings = (
            "Downloading golang dependencies. This may take a while...",
            "Finished downloading golang dependencies.",
            "Starting libp2p node...",
            "Connecting to libp2p node...",
            "Successfully connected to libp2p node!",
            LIBP2P_SUCCESS_MESSAGE,
        )
        missing_strings = self.missing_from_output(
            seller_aea_process, check_strings, timeout=240, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in seller_aea output.".format(missing_strings)

        self.set_agent_context(buyer_aea_name)
        buyer_aea_process = self.run_agent()

        check_strings = (
            "Downloading golang dependencies. This may take a while...",
            "Finished downloading golang dependencies.",
            "Starting libp2p node...",
            "Connecting to libp2p node...",
            "Successfully connected to libp2p node!",
            LIBP2P_SUCCESS_MESSAGE,
        )
        missing_strings = self.missing_from_output(
            buyer_aea_process, check_strings, timeout=240, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in buyer_aea output.".format(missing_strings)

        check_strings = (
            "registering agent on SOEF.",
            "registering service on SOEF.",
            "received CFP from sender=",
            "sending a PROPOSE with proposal=",
            "received ACCEPT from sender=",
            "sending MATCH_ACCEPT_W_INFORM to sender=",
            "received INFORM from sender=",
            "transaction confirmed, sending data=",
        )
        missing_strings = self.missing_from_output(
            seller_aea_process, check_strings, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in seller_aea output.".format(missing_strings)

        check_strings = (
            "found agents=",
            "sending CFP to agent=",
            "received proposal=",
            "accepting the proposal from sender=",
            "received MATCH_ACCEPT_W_INFORM from sender=",
            "informing counterparty=",
            "received INFORM from sender=",
            "received the following data=",
        )
        missing_strings = self.missing_from_output(
            buyer_aea_process, check_strings, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in buyer_aea output.".format(missing_strings)

        self.terminate_agents(seller_aea_process, buyer_aea_process)
        assert (
            self.is_successfully_terminated()
        ), "Agents weren't successfully terminated."
        wait_for_localhost_ports_to_close([9000, 9001])


@pytest.mark.sync
@pytest.mark.integration
class TestGenericSkillsFetchaiLedger(AEATestCaseMany):
    """Test that generic skills work."""

    @pytest.mark.flaky(
        reruns=MAX_FLAKY_RERUNS_INTEGRATION
    )  # cause possible network issues
    def test_generic(self, pytestconfig):
        """Run the generic skills sequence."""
        seller_aea_name = "my_generic_seller"
        buyer_aea_name = "my_generic_buyer"
        self.create_agents(seller_aea_name, buyer_aea_name)

        default_routing = {
            "fetchai/ledger_api:0.5.0": "fetchai/ledger:0.7.0",
            "fetchai/oef_search:0.8.0": "fetchai/soef:0.10.0",
        }

        # generate random location
        location = {
            "latitude": round(uniform(-90, 90), 2),  # nosec
            "longitude": round(uniform(-180, 180), 2),  # nosec
        }

        # prepare seller agent
        self.set_agent_context(seller_aea_name)
        self.add_item("connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/soef:0.10.0")
        self.set_config("agent.default_connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/ledger:0.7.0")
        self.add_item("skill", "fetchai/generic_seller:0.14.0")
        setting_path = "agent.default_routing"
        self.nested_set_config(setting_path, default_routing)
        self.run_install()

        diff = self.difference_to_fetched_agent(
            "fetchai/generic_seller:0.11.0", seller_aea_name
        )
        assert (
            diff == []
        ), "Difference between created and fetched project for files={}".format(diff)

        # add keys
        self.generate_private_key(FETCHAI)
        self.generate_private_key(COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION)
        self.add_private_key(FETCHAI, FETCHAI_PRIVATE_KEY_FILE)
        self.add_private_key(
            COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION, connection=True
        )
        self.replace_private_key_in_file(
            NON_FUNDED_COSMOS_PRIVATE_KEY_1, COSMOS_PRIVATE_KEY_FILE_CONNECTION
        )

        setting_path = "vendor.fetchai.connections.p2p_libp2p.config.ledger_id"
        self.set_config(setting_path, COSMOS)

        # make runable:
        setting_path = "vendor.fetchai.skills.generic_seller.is_abstract"
        self.set_config(setting_path, False, "bool")

        # replace location
        setting_path = (
            "vendor.fetchai.skills.generic_seller.models.strategy.args.location"
        )
        self.nested_set_config(setting_path, location)

        # prepare buyer agent
        self.set_agent_context(buyer_aea_name)
        self.add_item("connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/soef:0.10.0")
        self.set_config("agent.default_connection", "fetchai/p2p_libp2p:0.11.0")
        self.add_item("connection", "fetchai/ledger:0.7.0")
        self.add_item("skill", "fetchai/generic_buyer:0.13.0")
        setting_path = "agent.default_routing"
        self.nested_set_config(setting_path, default_routing)
        self.run_install()

        diff = self.difference_to_fetched_agent(
            "fetchai/generic_buyer:0.11.0", buyer_aea_name
        )
        assert (
            diff == []
        ), "Difference between created and fetched project for files={}".format(diff)

        setting_path = "vendor.fetchai.skills.generic_buyer.is_abstract"
        self.set_config(setting_path, False, "bool")

        # add keys
        self.generate_private_key(FETCHAI)
        self.generate_private_key(COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION)
        self.add_private_key(FETCHAI, FETCHAI_PRIVATE_KEY_FILE)
        self.add_private_key(
            COSMOS, COSMOS_PRIVATE_KEY_FILE_CONNECTION, connection=True
        )

        # fund key
        self.generate_wealth(FETCHAI)

        # set p2p configs
        setting_path = "vendor.fetchai.connections.p2p_libp2p.config"
        self.nested_set_config(setting_path, NON_GENESIS_CONFIG)

        # replace location
        setting_path = (
            "vendor.fetchai.skills.generic_buyer.models.strategy.args.location"
        )
        self.nested_set_config(setting_path, location)

        # run AEAs
        self.set_agent_context(seller_aea_name)
        seller_aea_process = self.run_agent()

        check_strings = (
            "Downloading golang dependencies. This may take a while...",
            "Finished downloading golang dependencies.",
            "Starting libp2p node...",
            "Connecting to libp2p node...",
            "Successfully connected to libp2p node!",
            LIBP2P_SUCCESS_MESSAGE,
        )
        missing_strings = self.missing_from_output(
            seller_aea_process, check_strings, timeout=240, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in seller_aea output.".format(missing_strings)

        self.set_agent_context(buyer_aea_name)
        buyer_aea_process = self.run_agent()

        check_strings = (
            "Downloading golang dependencies. This may take a while...",
            "Finished downloading golang dependencies.",
            "Starting libp2p node...",
            "Connecting to libp2p node...",
            "Successfully connected to libp2p node!",
            LIBP2P_SUCCESS_MESSAGE,
        )
        missing_strings = self.missing_from_output(
            buyer_aea_process, check_strings, timeout=240, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in buyer_aea output.".format(missing_strings)

        check_strings = (
            "registering agent on SOEF.",
            "registering service on SOEF.",
            "received CFP from sender=",
            "sending a PROPOSE with proposal=",
            "received ACCEPT from sender=",
            "sending MATCH_ACCEPT_W_INFORM to sender=",
            "received INFORM from sender=",
            "checking whether transaction=",
            "transaction confirmed, sending data=",
        )
        missing_strings = self.missing_from_output(
            seller_aea_process, check_strings, timeout=240, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in seller_aea output.".format(missing_strings)

        check_strings = (
            "found agents=",
            "sending CFP to agent=",
            "received proposal=",
            "accepting the proposal from sender=",
            "received MATCH_ACCEPT_W_INFORM from sender=",
            "requesting transfer transaction from ledger api...",
            "received raw transaction=",
            "proposing the transaction to the decision maker. Waiting for confirmation ...",
            "transaction signing was successful.",
            "sending transaction to ledger.",
            "transaction was successfully submitted. Transaction digest=",
            "informing counterparty=",
            "received INFORM from sender=",
            "received the following data=",
        )
        missing_strings = self.missing_from_output(
            buyer_aea_process, check_strings, is_terminating=False
        )
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in buyer_aea output.".format(missing_strings)

        self.terminate_agents(seller_aea_process, buyer_aea_process)
        assert (
            self.is_successfully_terminated()
        ), "Agents weren't successfully terminated."
        wait_for_localhost_ports_to_close([9000, 9001])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the p2p stub connection."""

import os
import tempfile
from pathlib import Path
from typing import Union, cast

from aea.configurations.base import ConnectionConfig, PublicId
from aea.connections.stub.connection import StubConnection, write_envelope
from aea.identity.base import Identity
from aea.mail.base import Envelope


PUBLIC_ID = PublicId.from_str("fetchai/p2p_stub:0.8.0")


class P2PStubConnection(StubConnection):
    r"""A p2p stub connection.

    This connection uses an existing directory as a Rendez-Vous point for agents to communicate locally.
    Each connected agent will create a file named after its address/identity where it can receive messages.

    The connection detects new messages by watchdogging the input file looking for new lines.
    """

    connection_id = PUBLIC_ID

    def __init__(self, configuration: ConnectionConfig, identity: Identity, **kwargs):
        """
        Initialize a p2p stub connection.

        :param configuration: the connection configuration
        :param identity: the identity
        """
        namespace_dir_path = cast(
            Union[str, Path],
            configuration.config.get("namespace_dir", tempfile.mkdtemp()),
        )
        if namespace_dir_path is None:
            raise ValueError("namespace_dir_path must be set!")  # pragma: nocover
        self.namespace = os.path.abspath(namespace_dir_path)

        input_file_path = os.path.join(self.namespace, "{}.in".format(identity.address))
        output_file_path = os.path.join(
            self.namespace, "{}.out".format(identity.address)
        )
        configuration.config["input_file"] = input_file_path
        configuration.config["output_file"] = output_file_path
        super().__init__(configuration=configuration, identity=identity, **kwargs)

    async def send(self, envelope: Envelope):
        """
        Send messages.

        :return: None
        """
        if self.loop is None:
            raise ValueError("Loop not initialized.")  # pragma: nocover
        self._ensure_valid_envelope_for_external_comms(envelope)
        target_file = Path(os.path.join(self.namespace, "{}.in".format(envelope.to)))

        with open(target_file, "ab") as file:
            await self.loop.run_in_executor(
                self._write_pool, write_envelope, envelope, file
            )

    async def disconnect(self) -> None:
        """Disconnect the connection."""
        if self.loop is None:
            raise ValueError("Loop not initialized.")  # pragma: nocover
        await self.loop.run_in_executor(self._write_pool, self._cleanup)
        await super().disconnect()

    def _cleanup(self):
        try:
            os.unlink(self.configuration.config["input_file"])
        except OSError:
            pass
        try:
            os.unlink(self.configuration.config["output_file"])
        except OSError:
            pass
        try:
            os.rmdir(self.namespace)
        except OSError:
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains t_protocol_no_ct's message definition."""

import logging
from typing import Dict, FrozenSet, Optional, Set, Tuple, Union, cast

from aea.configurations.base import ProtocolId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message


_default_logger = logging.getLogger(
    "aea.packages.fetchai.protocols.t_protocol_no_ct.message"
)

DEFAULT_BODY_SIZE = 4


class TProtocolNoCtMessage(Message):
    """A protocol for testing purposes."""

    protocol_id = ProtocolId.from_str("fetchai/t_protocol_no_ct:0.1.0")

    class Performative(Message.Performative):
        """Performatives for the t_protocol_no_ct protocol."""

        PERFORMATIVE_EMPTY_CONTENTS = "performative_empty_contents"
        PERFORMATIVE_MT = "performative_mt"
        PERFORMATIVE_O = "performative_o"
        PERFORMATIVE_PCT = "performative_pct"
        PERFORMATIVE_PMT = "performative_pmt"
        PERFORMATIVE_PT = "performative_pt"

        def __str__(self):
            """Get the string representation."""
            return str(self.value)

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs,
    ):
        """
        Initialise an instance of TProtocolNoCtMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        """
        self._performatives = {
            "performative_empty_contents",
            "performative_mt",
            "performative_o",
            "performative_pct",
            "performative_pmt",
            "performative_pt",
        }
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=TProtocolNoCtMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(TProtocolNoCtMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def content_bool(self) -> bool:
        """Get the 'content_bool' content from the message."""
        enforce(self.is_set("content_bool"), "'content_bool' content is not set.")
        return cast(bool, self.get("content_bool"))

    @property
    def content_bytes(self) -> bytes:
        """Get the 'content_bytes' content from the message."""
        enforce(self.is_set("content_bytes"), "'content_bytes' content is not set.")
        return cast(bytes, self.get("content_bytes"))

    @property
    def content_dict_bool_bool(self) -> Dict[bool, bool]:
        """Get the 'content_dict_bool_bool' content from the message."""
        enforce(
            self.is_set("content_dict_bool_bool"),
            "'content_dict_bool_bool' content is not set.",
        )
        return cast(Dict[bool, bool], self.get("content_dict_bool_bool"))

    @property
    def content_dict_bool_bytes(self) -> Dict[bool, bytes]:
        """Get the 'content_dict_bool_bytes' content from the message."""
        enforce(
            self.is_set("content_dict_bool_bytes"),
            "'content_dict_bool_bytes' content is not set.",
        )
        return cast(Dict[bool, bytes], self.get("content_dict_bool_bytes"))

    @property
    def content_dict_bool_float(self) -> Dict[bool, float]:
        """Get the 'content_dict_bool_float' content from the message."""
        enforce(
            self.is_set("content_dict_bool_float"),
            "'content_dict_bool_float' content is not set.",
        )
        return cast(Dict[bool, float], self.get("content_dict_bool_float"))

    @property
    def content_dict_bool_int(self) -> Dict[bool, int]:
        """Get the 'content_dict_bool_int' content from the message."""
        enforce(
            self.is_set("content_dict_bool_int"),
            "'content_dict_bool_int' content is not set.",
        )
        return cast(Dict[bool, int], self.get("content_dict_bool_int"))

    @property
    def content_dict_bool_str(self) -> Dict[bool, str]:
        """Get the 'content_dict_bool_str' content from the message."""
        enforce(
            self.is_set("content_dict_bool_str"),
            "'content_dict_bool_str' content is not set.",
        )
        return cast(Dict[bool, str], self.get("content_dict_bool_str"))

    @property
    def content_dict_int_bool(self) -> Dict[int, bool]:
        """Get the 'content_dict_int_bool' content from the message."""
        enforce(
            self.is_set("content_dict_int_bool"),
            "'content_dict_int_bool' content is not set.",
        )
        return cast(Dict[int, bool], self.get("content_dict_int_bool"))

    @property
    def content_dict_int_bytes(self) -> Dict[int, bytes]:
        """Get the 'content_dict_int_bytes' content from the message."""
        enforce(
            self.is_set("content_dict_int_bytes"),
            "'content_dict_int_bytes' content is not set.",
        )
        return cast(Dict[int, bytes], self.get("content_dict_int_bytes"))

    @property
    def content_dict_int_float(self) -> Dict[int, float]:
        """Get the 'content_dict_int_float' content from the message."""
        enforce(
            self.is_set("content_dict_int_float"),
            "'content_dict_int_float' content is not set.",
        )
        return cast(Dict[int, float], self.get("content_dict_int_float"))

    @property
    def content_dict_int_int(self) -> Dict[int, int]:
        """Get the 'content_dict_int_int' content from the message."""
        enforce(
            self.is_set("content_dict_int_int"),
            "'content_dict_int_int' content is not set.",
        )
        return cast(Dict[int, int], self.get("content_dict_int_int"))

    @property
    def content_dict_int_str(self) -> Dict[int, str]:
        """Get the 'content_dict_int_str' content from the message."""
        enforce(
            self.is_set("content_dict_int_str"),
            "'content_dict_int_str' content is not set.",
        )
        return cast(Dict[int, str], self.get("content_dict_int_str"))

    @property
    def content_dict_str_bool(self) -> Dict[str, bool]:
        """Get the 'content_dict_str_bool' content from the message."""
        enforce(
            self.is_set("content_dict_str_bool"),
            "'content_dict_str_bool' content is not set.",
        )
        return cast(Dict[str, bool], self.get("content_dict_str_bool"))

    @property
    def content_dict_str_bytes(self) -> Dict[str, bytes]:
        """Get the 'content_dict_str_bytes' content from the message."""
        enforce(
            self.is_set("content_dict_str_bytes"),
            "'content_dict_str_bytes' content is not set.",
        )
        return cast(Dict[str, bytes], self.get("content_dict_str_bytes"))

    @property
    def content_dict_str_float(self) -> Dict[str, float]:
        """Get the 'content_dict_str_float' content from the message."""
        enforce(
            self.is_set("content_dict_str_float"),
            "'content_dict_str_float' content is not set.",
        )
        return cast(Dict[str, float], self.get("content_dict_str_float"))

    @property
    def content_dict_str_int(self) -> Dict[str, int]:
        """Get the 'content_dict_str_int' content from the message."""
        enforce(
            self.is_set("content_dict_str_int"),
            "'content_dict_str_int' content is not set.",
        )
        return cast(Dict[str, int], self.get("content_dict_str_int"))

    @property
    def content_dict_str_str(self) -> Dict[str, str]:
        """Get the 'content_dict_str_str' content from the message."""
        enforce(
            self.is_set("content_dict_str_str"),
            "'content_dict_str_str' content is not set.",
        )
        return cast(Dict[str, str], self.get("content_dict_str_str"))

    @property
    def content_float(self) -> float:
        """Get the 'content_float' content from the message."""
        enforce(self.is_set("content_float"), "'content_float' content is not set.")
        return cast(float, self.get("content_float"))

    @property
    def content_int(self) -> int:
        """Get the 'content_int' content from the message."""
        enforce(self.is_set("content_int"), "'content_int' content is not set.")
        return cast(int, self.get("content_int"))

    @property
    def content_list_bool(self) -> Tuple[bool, ...]:
        """Get the 'content_list_bool' content from the message."""
        enforce(
            self.is_set("content_list_bool"), "'content_list_bool' content is not set."
        )
        return cast(Tuple[bool, ...], self.get("content_list_bool"))

    @property
    def content_list_bytes(self) -> Tuple[bytes, ...]:
        """Get the 'content_list_bytes' content from the message."""
        enforce(
            self.is_set("content_list_bytes"),
            "'content_list_bytes' content is not set.",
        )
        return cast(Tuple[bytes, ...], self.get("content_list_bytes"))

    @property
    def content_list_float(self) -> Tuple[float, ...]:
        """Get the 'content_list_float' content from the message."""
        enforce(
            self.is_set("content_list_float"),
            "'content_list_float' content is not set.",
        )
        return cast(Tuple[float, ...], self.get("content_list_float"))

    @property
    def content_list_int(self) -> Tuple[int, ...]:
        """Get the 'content_list_int' content from the message."""
        enforce(
            self.is_set("content_list_int"), "'content_list_int' content is not set."
        )
        return cast(Tuple[int, ...], self.get("content_list_int"))

    @property
    def content_list_str(self) -> Tuple[str, ...]:
        """Get the 'content_list_str' content from the message."""
        enforce(
            self.is_set("content_list_str"), "'content_list_str' content is not set."
        )
        return cast(Tuple[str, ...], self.get("content_list_str"))

    @property
    def content_o_bool(self) -> Optional[bool]:
        """Get the 'content_o_bool' content from the message."""
        return cast(Optional[bool], self.get("content_o_bool"))

    @property
    def content_o_dict_str_int(self) -> Optional[Dict[str, int]]:
        """Get the 'content_o_dict_str_int' content from the message."""
        return cast(Optional[Dict[str, int]], self.get("content_o_dict_str_int"))

    @property
    def content_o_list_bytes(self) -> Optional[Tuple[bytes, ...]]:
        """Get the 'content_o_list_bytes' content from the message."""
        return cast(Optional[Tuple[bytes, ...]], self.get("content_o_list_bytes"))

    @property
    def content_o_set_int(self) -> Optional[FrozenSet[int]]:
        """Get the 'content_o_set_int' content from the message."""
        return cast(Optional[FrozenSet[int]], self.get("content_o_set_int"))

    @property
    def content_set_bool(self) -> FrozenSet[bool]:
        """Get the 'content_set_bool' content from the message."""
        enforce(
            self.is_set("content_set_bool"), "'content_set_bool' content is not set."
        )
        return cast(FrozenSet[bool], self.get("content_set_bool"))

    @property
    def content_set_bytes(self) -> FrozenSet[bytes]:
        """Get the 'content_set_bytes' content from the message."""
        enforce(
            self.is_set("content_set_bytes"), "'content_set_bytes' content is not set."
        )
        return cast(FrozenSet[bytes], self.get("content_set_bytes"))

    @property
    def content_set_float(self) -> FrozenSet[float]:
        """Get the 'content_set_float' content from the message."""
        enforce(
            self.is_set("content_set_float"), "'content_set_float' content is not set."
        )
        return cast(FrozenSet[float], self.get("content_set_float"))

    @property
    def content_set_int(self) -> FrozenSet[int]:
        """Get the 'content_set_int' content from the message."""
        enforce(self.is_set("content_set_int"), "'content_set_int' content is not set.")
        return cast(FrozenSet[int], self.get("content_set_int"))

    @property
    def content_set_str(self) -> FrozenSet[str]:
        """Get the 'content_set_str' content from the message."""
        enforce(self.is_set("content_set_str"), "'content_set_str' content is not set.")
        return cast(FrozenSet[str], self.get("content_set_str"))

    @property
    def content_str(self) -> str:
        """Get the 'content_str' content from the message."""
        enforce(self.is_set("content_str"), "'content_str' content is not set.")
        return cast(str, self.get("content_str"))

    @property
    def content_union_1(
        self,
    ) -> Union[
        bytes, int, float, bool, str, FrozenSet[int], Tuple[bool, ...], Dict[str, int]
    ]:
        """Get the 'content_union_1' content from the message."""
        enforce(self.is_set("content_union_1"), "'content_union_1' content is not set.")
        return cast(
            Union[
                bytes,
                int,
                float,
                bool,
                str,
                FrozenSet[int],
                Tuple[bool, ...],
                Dict[str, int],
            ],
            self.get("content_union_1"),
        )

    @property
    def content_union_2(
        self,
    ) -> Union[
        FrozenSet[bytes],
        FrozenSet[int],
        FrozenSet[str],
        Tuple[float, ...],
        Tuple[bool, ...],
        Tuple[bytes, ...],
        Dict[str, int],
        Dict[int, float],
        Dict[bool, bytes],
    ]:
        """Get the 'content_union_2' content from the message."""
        enforce(self.is_set("content_union_2"), "'content_union_2' content is not set.")
        return cast(
            Union[
                FrozenSet[bytes],
                FrozenSet[int],
                FrozenSet[str],
                Tuple[float, ...],
                Tuple[bool, ...],
                Tuple[bytes, ...],
                Dict[str, int],
                Dict[int, float],
                Dict[bool, bytes],
            ],
            self.get("content_union_2"),
        )

    def _is_consistent(self) -> bool:
        """Check that the message follows the t_protocol_no_ct protocol."""
        try:
            enforce(
                type(self.dialogue_reference) == tuple,
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                type(self.dialogue_reference[0]) == str,
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                type(self.dialogue_reference[1]) == str,
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) == int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) == int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                type(self.performative) == TProtocolNoCtMessage.Performative,
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self.body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == TProtocolNoCtMessage.Performative.PERFORMATIVE_PT:
                expected_nb_of_contents = 5
                enforce(
                    type(self.content_bytes) == bytes,
                    "Invalid type for content 'content_bytes'. Expected 'bytes'. Found '{}'.".format(
                        type(self.content_bytes)
                    ),
                )
                enforce(
                    type(self.content_int) == int,
                    "Invalid type for content 'content_int'. Expected 'int'. Found '{}'.".format(
                        type(self.content_int)
                    ),
                )
                enforce(
                    type(self.content_float) == float,
                    "Invalid type for content 'content_float'. Expected 'float'. Found '{}'.".format(
                        type(self.content_float)
                    ),
                )
                enforce(
                    type(self.content_bool) == bool,
                    "Invalid type for content 'content_bool'. Expected 'bool'. Found '{}'.".format(
                        type(self.content_bool)
                    ),
                )
                enforce(
                    type(self.content_str) == str,
                    "Invalid type for content 'content_str'. Expected 'str'. Found '{}'.".format(
                        type(self.content_str)
                    ),
                )
            elif (
                self.performative == TProtocolNoCtMessage.Performative.PERFORMATIVE_PCT
            ):
                expected_nb_of_contents = 10
                enforce(
                    type(self.content_set_bytes) == frozenset,
                    "Invalid type for content 'content_set_bytes'. Expected 'frozenset'. Found '{}'.".format(
                        type(self.content_set_bytes)
                    ),
                )
                enforce(
                    all(type(element) == bytes for element in self.content_set_bytes),
                    "Invalid type for frozenset elements in content 'content_set_bytes'. Expected 'bytes'.",
                )
                enforce(
                    type(self.content_set_int) == frozenset,
                    "Invalid type for content 'content_set_int'. Expected 'frozenset'. Found '{}'.".format(
                        type(self.content_set_int)
                    ),
                )
                enforce(
                    all(type(element) == int for element in self.content_set_int),
                    "Invalid type for frozenset elements in content 'content_set_int'. Expected 'int'.",
                )
                enforce(
                    type(self.content_set_float) == frozenset,
                    "Invalid type for content 'content_set_float'. Expected 'frozenset'. Found '{}'.".format(
                        type(self.content_set_float)
                    ),
                )
                enforce(
                    all(type(element) == float for element in self.content_set_float),
                    "Invalid type for frozenset elements in content 'content_set_float'. Expected 'float'.",
                )
                enforce(
                    type(self.content_set_bool) == frozenset,
                    "Invalid type for content 'content_set_bool'. Expected 'frozenset'. Found '{}'.".format(
                        type(self.content_set_bool)
                    ),
                )
                enforce(
                    all(type(element) == bool for element in self.content_set_bool),
                    "Invalid type for frozenset elements in content 'content_set_bool'. Expected 'bool'.",
                )
                enforce(
                    type(self.content_set_str) == frozenset,
                    "Invalid type for content 'content_set_str'. Expected 'frozenset'. Found '{}'.".format(
                        type(self.content_set_str)
                    ),
                )
                enforce(
                    all(type(element) == str for element in self.content_set_str),
                    "Invalid type for frozenset elements in content 'content_set_str'. Expected 'str'.",
                )
                enforce(
                    type(self.content_list_bytes) == tuple,
                    "Invalid type for content 'content_list_bytes'. Expected 'tuple'. Found '{}'.".format(
                        type(self.content_list_bytes)
                    ),
                )
                enforce(
                    all(type(element) == bytes for element in self.content_list_bytes),
                    "Invalid type for tuple elements in content 'content_list_bytes'. Expected 'bytes'.",
                )
                enforce(
                    type(self.content_list_int) == tuple,
                    "Invalid type for content 'content_list_int'. Expected 'tuple'. Found '{}'.".format(
                        type(self.content_list_int)
                    ),
                )
                enforce(
                    all(type(element) == int for element in self.content_list_int),
                    "Invalid type for tuple elements in content 'content_list_int'. Expected 'int'.",
                )
                enforce(
                    type(self.content_list_float) == tuple,
                    "Invalid type for content 'content_list_float'. Expected 'tuple'. Found '{}'.".format(
                        type(self.content_list_float)
                    ),
                )
                enforce(
                    all(type(element) == float for element in self.content_list_float),
                    "Invalid type for tuple elements in content 'content_list_float'. Expected 'float'.",
                )
                enforce(
                    type(self.content_list_bool) == tuple,
                    "Invalid type for content 'content_list_bool'. Expected 'tuple'. Found '{}'.".format(
                        type(self.content_list_bool)
                    ),
                )
                enforce(
                    all(type(element) == bool for element in self.content_list_bool),
                    "Invalid type for tuple elements in content 'content_list_bool'. Expected 'bool'.",
                )
                enforce(
                    type(self.content_list_str) == tuple,
                    "Invalid type for content 'content_list_str'. Expected 'tuple'. Found '{}'.".format(
                        type(self.content_list_str)
                    ),
                )
                enforce(
                    all(type(element) == str for element in self.content_list_str),
                    "Invalid type for tuple elements in content 'content_list_str'. Expected 'str'.",
                )
            elif (
                self.performative == TProtocolNoCtMessage.Performative.PERFORMATIVE_PMT
            ):
                expected_nb_of_contents = 15
                enforce(
                    type(self.content_dict_int_bytes) == dict,
                    "Invalid type for content 'content_dict_int_bytes'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_int_bytes)
                    ),
                )
                for (
                    key_of_content_dict_int_bytes,
                    value_of_content_dict_int_bytes,
                ) in self.content_dict_int_bytes.items():
                    enforce(
                        type(key_of_content_dict_int_bytes) == int,
                        "Invalid type for dictionary keys in content 'content_dict_int_bytes'. Expected 'int'. Found '{}'.".format(
                            type(key_of_content_dict_int_bytes)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_int_bytes) == bytes,
                        "Invalid type for dictionary values in content 'content_dict_int_bytes'. Expected 'bytes'. Found '{}'.".format(
                            type(value_of_content_dict_int_bytes)
                        ),
                    )
                enforce(
                    type(self.content_dict_int_int) == dict,
                    "Invalid type for content 'content_dict_int_int'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_int_int)
                    ),
                )
                for (
                    key_of_content_dict_int_int,
                    value_of_content_dict_int_int,
                ) in self.content_dict_int_int.items():
                    enforce(
                        type(key_of_content_dict_int_int) == int,
                        "Invalid type for dictionary keys in content 'content_dict_int_int'. Expected 'int'. Found '{}'.".format(
                            type(key_of_content_dict_int_int)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_int_int) == int,
                        "Invalid type for dictionary values in content 'content_dict_int_int'. Expected 'int'. Found '{}'.".format(
                            type(value_of_content_dict_int_int)
                        ),
                    )
                enforce(
                    type(self.content_dict_int_float) == dict,
                    "Invalid type for content 'content_dict_int_float'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_int_float)
                    ),
                )
                for (
                    key_of_content_dict_int_float,
                    value_of_content_dict_int_float,
                ) in self.content_dict_int_float.items():
                    enforce(
                        type(key_of_content_dict_int_float) == int,
                        "Invalid type for dictionary keys in content 'content_dict_int_float'. Expected 'int'. Found '{}'.".format(
                            type(key_of_content_dict_int_float)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_int_float) == float,
                        "Invalid type for dictionary values in content 'content_dict_int_float'. Expected 'float'. Found '{}'.".format(
                            type(value_of_content_dict_int_float)
                        ),
                    )
                enforce(
                    type(self.content_dict_int_bool) == dict,
                    "Invalid type for content 'content_dict_int_bool'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_int_bool)
                    ),
                )
                for (
                    key_of_content_dict_int_bool,
                    value_of_content_dict_int_bool,
                ) in self.content_dict_int_bool.items():
                    enforce(
                        type(key_of_content_dict_int_bool) == int,
                        "Invalid type for dictionary keys in content 'content_dict_int_bool'. Expected 'int'. Found '{}'.".format(
                            type(key_of_content_dict_int_bool)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_int_bool) == bool,
                        "Invalid type for dictionary values in content 'content_dict_int_bool'. Expected 'bool'. Found '{}'.".format(
                            type(value_of_content_dict_int_bool)
                        ),
                    )
                enforce(
                    type(self.content_dict_int_str) == dict,
                    "Invalid type for content 'content_dict_int_str'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_int_str)
                    ),
                )
                for (
                    key_of_content_dict_int_str,
                    value_of_content_dict_int_str,
                ) in self.content_dict_int_str.items():
                    enforce(
                        type(key_of_content_dict_int_str) == int,
                        "Invalid type for dictionary keys in content 'content_dict_int_str'. Expected 'int'. Found '{}'.".format(
                            type(key_of_content_dict_int_str)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_int_str) == str,
                        "Invalid type for dictionary values in content 'content_dict_int_str'. Expected 'str'. Found '{}'.".format(
                            type(value_of_content_dict_int_str)
                        ),
                    )
                enforce(
                    type(self.content_dict_bool_bytes) == dict,
                    "Invalid type for content 'content_dict_bool_bytes'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_bool_bytes)
                    ),
                )
                for (
                    key_of_content_dict_bool_bytes,
                    value_of_content_dict_bool_bytes,
                ) in self.content_dict_bool_bytes.items():
                    enforce(
                        type(key_of_content_dict_bool_bytes) == bool,
                        "Invalid type for dictionary keys in content 'content_dict_bool_bytes'. Expected 'bool'. Found '{}'.".format(
                            type(key_of_content_dict_bool_bytes)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_bool_bytes) == bytes,
                        "Invalid type for dictionary values in content 'content_dict_bool_bytes'. Expected 'bytes'. Found '{}'.".format(
                            type(value_of_content_dict_bool_bytes)
                        ),
                    )
                enforce(
                    type(self.content_dict_bool_int) == dict,
                    "Invalid type for content 'content_dict_bool_int'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_bool_int)
                    ),
                )
                for (
                    key_of_content_dict_bool_int,
                    value_of_content_dict_bool_int,
                ) in self.content_dict_bool_int.items():
                    enforce(
                        type(key_of_content_dict_bool_int) == bool,
                        "Invalid type for dictionary keys in content 'content_dict_bool_int'. Expected 'bool'. Found '{}'.".format(
                            type(key_of_content_dict_bool_int)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_bool_int) == int,
                        "Invalid type for dictionary values in content 'content_dict_bool_int'. Expected 'int'. Found '{}'.".format(
                            type(value_of_content_dict_bool_int)
                        ),
                    )
                enforce(
                    type(self.content_dict_bool_float) == dict,
                    "Invalid type for content 'content_dict_bool_float'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_bool_float)
                    ),
                )
                for (
                    key_of_content_dict_bool_float,
                    value_of_content_dict_bool_float,
                ) in self.content_dict_bool_float.items():
                    enforce(
                        type(key_of_content_dict_bool_float) == bool,
                        "Invalid type for dictionary keys in content 'content_dict_bool_float'. Expected 'bool'. Found '{}'.".format(
                            type(key_of_content_dict_bool_float)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_bool_float) == float,
                        "Invalid type for dictionary values in content 'content_dict_bool_float'. Expected 'float'. Found '{}'.".format(
                            type(value_of_content_dict_bool_float)
                        ),
                    )
                enforce(
                    type(self.content_dict_bool_bool) == dict,
                    "Invalid type for content 'content_dict_bool_bool'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_bool_bool)
                    ),
                )
                for (
                    key_of_content_dict_bool_bool,
                    value_of_content_dict_bool_bool,
                ) in self.content_dict_bool_bool.items():
                    enforce(
                        type(key_of_content_dict_bool_bool) == bool,
                        "Invalid type for dictionary keys in content 'content_dict_bool_bool'. Expected 'bool'. Found '{}'.".format(
                            type(key_of_content_dict_bool_bool)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_bool_bool) == bool,
                        "Invalid type for dictionary values in content 'content_dict_bool_bool'. Expected 'bool'. Found '{}'.".format(
                            type(value_of_content_dict_bool_bool)
                        ),
                    )
                enforce(
                    type(self.content_dict_bool_str) == dict,
                    "Invalid type for content 'content_dict_bool_str'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_bool_str)
                    ),
                )
                for (
                    key_of_content_dict_bool_str,
                    value_of_content_dict_bool_str,
                ) in self.content_dict_bool_str.items():
                    enforce(
                        type(key_of_content_dict_bool_str) == bool,
                        "Invalid type for dictionary keys in content 'content_dict_bool_str'. Expected 'bool'. Found '{}'.".format(
                            type(key_of_content_dict_bool_str)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_bool_str) == str,
                        "Invalid type for dictionary values in content 'content_dict_bool_str'. Expected 'str'. Found '{}'.".format(
                            type(value_of_content_dict_bool_str)
                        ),
                    )
                enforce(
                    type(self.content_dict_str_bytes) == dict,
                    "Invalid type for content 'content_dict_str_bytes'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_str_bytes)
                    ),
                )
                for (
                    key_of_content_dict_str_bytes,
                    value_of_content_dict_str_bytes,
                ) in self.content_dict_str_bytes.items():
                    enforce(
                        type(key_of_content_dict_str_bytes) == str,
                        "Invalid type for dictionary keys in content 'content_dict_str_bytes'. Expected 'str'. Found '{}'.".format(
                            type(key_of_content_dict_str_bytes)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_str_bytes) == bytes,
                        "Invalid type for dictionary values in content 'content_dict_str_bytes'. Expected 'bytes'. Found '{}'.".format(
                            type(value_of_content_dict_str_bytes)
                        ),
                    )
                enforce(
                    type(self.content_dict_str_int) == dict,
                    "Invalid type for content 'content_dict_str_int'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_str_int)
                    ),
                )
                for (
                    key_of_content_dict_str_int,
                    value_of_content_dict_str_int,
                ) in self.content_dict_str_int.items():
                    enforce(
                        type(key_of_content_dict_str_int) == str,
                        "Invalid type for dictionary keys in content 'content_dict_str_int'. Expected 'str'. Found '{}'.".format(
                            type(key_of_content_dict_str_int)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_str_int) == int,
                        "Invalid type for dictionary values in content 'content_dict_str_int'. Expected 'int'. Found '{}'.".format(
                            type(value_of_content_dict_str_int)
                        ),
                    )
                enforce(
                    type(self.content_dict_str_float) == dict,
                    "Invalid type for content 'content_dict_str_float'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_str_float)
                    ),
                )
                for (
                    key_of_content_dict_str_float,
                    value_of_content_dict_str_float,
                ) in self.content_dict_str_float.items():
                    enforce(
                        type(key_of_content_dict_str_float) == str,
                        "Invalid type for dictionary keys in content 'content_dict_str_float'. Expected 'str'. Found '{}'.".format(
                            type(key_of_content_dict_str_float)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_str_float) == float,
                        "Invalid type for dictionary values in content 'content_dict_str_float'. Expected 'float'. Found '{}'.".format(
                            type(value_of_content_dict_str_float)
                        ),
                    )
                enforce(
                    type(self.content_dict_str_bool) == dict,
                    "Invalid type for content 'content_dict_str_bool'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_str_bool)
                    ),
                )
                for (
                    key_of_content_dict_str_bool,
                    value_of_content_dict_str_bool,
                ) in self.content_dict_str_bool.items():
                    enforce(
                        type(key_of_content_dict_str_bool) == str,
                        "Invalid type for dictionary keys in content 'content_dict_str_bool'. Expected 'str'. Found '{}'.".format(
                            type(key_of_content_dict_str_bool)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_str_bool) == bool,
                        "Invalid type for dictionary values in content 'content_dict_str_bool'. Expected 'bool'. Found '{}'.".format(
                            type(value_of_content_dict_str_bool)
                        ),
                    )
                enforce(
                    type(self.content_dict_str_str) == dict,
                    "Invalid type for content 'content_dict_str_str'. Expected 'dict'. Found '{}'.".format(
                        type(self.content_dict_str_str)
                    ),
                )
                for (
                    key_of_content_dict_str_str,
                    value_of_content_dict_str_str,
                ) in self.content_dict_str_str.items():
                    enforce(
                        type(key_of_content_dict_str_str) == str,
                        "Invalid type for dictionary keys in content 'content_dict_str_str'. Expected 'str'. Found '{}'.".format(
                            type(key_of_content_dict_str_str)
                        ),
                    )
                    enforce(
                        type(value_of_content_dict_str_str) == str,
                        "Invalid type for dictionary values in content 'content_dict_str_str'. Expected 'str'. Found '{}'.".format(
                            type(value_of_content_dict_str_str)
                        ),
                    )
            elif self.performative == TProtocolNoCtMessage.Performative.PERFORMATIVE_MT:
                expected_nb_of_contents = 2
                enforce(
                    type(self.content_union_1) == bool
                    or type(self.content_union_1) == bytes
                    or type(self.content_union_1) == dict
                    or type(self.content_union_1) == float
                    or type(self.content_union_1) == frozenset
                    or type(self.content_union_1) == int
                    or type(self.content_union_1) == str
                    or type(self.content_union_1) == tuple,
                    "Invalid type for content 'content_union_1'. Expected either of '['bool', 'bytes', 'dict', 'float', 'frozenset', 'int', 'str', 'tuple']'. Found '{}'.".format(
                        type(self.content_union_1)
                    ),
                )
                if type(self.content_union_1) == frozenset:
                    enforce(
                        all(type(element) == int for element in self.content_union_1),
                        "Invalid type for elements of content 'content_union_1'. Expected 'int'.",
                    )
                if type(self.content_union_1) == tuple:
                    enforce(
                        all(type(element) == bool for element in self.content_union_1),
                        "Invalid type for tuple elements in content 'content_union_1'. Expected 'bool'.",
                    )
                if type(self.content_union_1) == dict:
                    for (
                        key_of_content_union_1,
                        value_of_content_union_1,
                    ) in self.content_union_1.items():
                        enforce(
                            (
                                type(key_of_content_union_1) == str
                                and type(value_of_content_union_1) == int
                            ),
                            "Invalid type for dictionary key, value in content 'content_union_1'. Expected 'str', 'int'.",
                        )
                enforce(
                    type(self.content_union_2) == dict
                    or type(self.content_union_2) == frozenset
                    or type(self.content_union_2) == tuple,
                    "Invalid type for content 'content_union_2'. Expected either of '['dict', 'frozenset', 'tuple']'. Found '{}'.".format(
                        type(self.content_union_2)
                    ),
                )
                if type(self.content_union_2) == frozenset:
                    enforce(
                        all(type(element) == bytes for element in self.content_union_2)
                        or all(type(element) == int for element in self.content_union_2)
                        or all(
                            type(element) == str for element in self.content_union_2
                        ),
                        "Invalid type for frozenset elements in content 'content_union_2'. Expected either 'bytes' or 'int' or 'str'.",
                    )
                if type(self.content_union_2) == tuple:
                    enforce(
                        all(type(element) == bool for element in self.content_union_2)
                        or all(
                            type(element) == bytes for element in self.content_union_2
                        )
                        or all(
                            type(element) == float for element in self.content_union_2
                        ),
                        "Invalid type for tuple elements in content 'content_union_2'. Expected either 'bool' or 'bytes' or 'float'.",
                    )
                if type(self.content_union_2) == dict:
                    for (
                        key_of_content_union_2,
                        value_of_content_union_2,
                    ) in self.content_union_2.items():
                        enforce(
                            (
                                type(key_of_content_union_2) == bool
                                and type(value_of_content_union_2) == bytes
                            )
                            or (
                                type(key_of_content_union_2) == int
                                and type(value_of_content_union_2) == float
                            )
                            or (
                                type(key_of_content_union_2) == str
                                and type(value_of_content_union_2) == int
                            ),
                            "Invalid type for dictionary key, value in content 'content_union_2'. Expected 'bool','bytes' or 'int','float' or 'str','int'.",
                        )
            elif self.performative == TProtocolNoCtMessage.Performative.PERFORMATIVE_O:
                expected_nb_of_contents = 0
                if self.is_set("content_o_bool"):
                    expected_nb_of_contents += 1
                    content_o_bool = cast(bool, self.content_o_bool)
                    enforce(
                        type(content_o_bool) == bool,
                        "Invalid type for content 'content_o_bool'. Expected 'bool'. Found '{}'.".format(
                            type(content_o_bool)
                        ),
                    )
                if self.is_set("content_o_set_int"):
                    expected_nb_of_contents += 1
                    content_o_set_int = cast(FrozenSet[int], self.content_o_set_int)
                    enforce(
                        type(content_o_set_int) == frozenset,
                        "Invalid type for content 'content_o_set_int'. Expected 'frozenset'. Found '{}'.".format(
                            type(content_o_set_int)
                        ),
                    )
                    enforce(
                        all(type(element) == int for element in content_o_set_int),
                        "Invalid type for frozenset elements in content 'content_o_set_int'. Expected 'int'.",
                    )
                if self.is_set("content_o_list_bytes"):
                    expected_nb_of_contents += 1
                    content_o_list_bytes = cast(
                        Tuple[bytes, ...], self.content_o_list_bytes
                    )
                    enforce(
                        type(content_o_list_bytes) == tuple,
                        "Invalid type for content 'content_o_list_bytes'. Expected 'tuple'. Found '{}'.".format(
                            type(content_o_list_bytes)
                        ),
                    )
                    enforce(
                        all(type(element) == bytes for element in content_o_list_bytes),
                        "Invalid type for tuple elements in content 'content_o_list_bytes'. Expected 'bytes'.",
                    )
                if self.is_set("content_o_dict_str_int"):
                    expected_nb_of_contents += 1
                    content_o_dict_str_int = cast(
                        Dict[str, int], self.content_o_dict_str_int
                    )
                    enforce(
                        type(content_o_dict_str_int) == dict,
                        "Invalid type for content 'content_o_dict_str_int'. Expected 'dict'. Found '{}'.".format(
                            type(content_o_dict_str_int)
                        ),
                    )
                    for (
                        key_of_content_o_dict_str_int,
                        value_of_content_o_dict_str_int,
                    ) in content_o_dict_str_int.items():
                        enforce(
                            type(key_of_content_o_dict_str_int) == str,
                            "Invalid type for dictionary keys in content 'content_o_dict_str_int'. Expected 'str'. Found '{}'.".format(
                                type(key_of_content_o_dict_str_int)
                            ),
                        )
                        enforce(
                            type(value_of_content_o_dict_str_int) == int,
                            "Invalid type for dictionary values in content 'content_o_dict_str_int'. Expected 'int'. Found '{}'.".format(
                                type(value_of_content_o_dict_str_int)
                            ),
                        )
            elif (
                self.performative
                == TProtocolNoCtMessage.Performative.PERFORMATIVE_EMPTY_CONTENTS
            ):
                expected_nb_of_contents = 0

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
            else:
                enforce(
                    0 < self.target < self.message_id,
                    "Invalid 'target'. Expected an integer between 1 and {} inclusive. Found {}.".format(
                        self.message_id - 1, self.target,
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the protocol.md file."""
from enum import Enum
from pathlib import Path

import mistune

from aea.protocols.default.message import DefaultMessage

from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.custom_types import OefErrorOperation
from packages.fetchai.protocols.oef_search.message import OefSearchMessage

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import compare_enum_classes, compile_and_exec


class TestProtocolDocs:
    """Test the integrity of the code-blocks in skill.md"""

    @classmethod
    def setup_class(cls):
        """Test skill.md"""
        markdown_parser = mistune.create_markdown(renderer=mistune.AstRenderer())

        skill_doc_file = Path(ROOT_DIR, "docs", "protocol.md")
        doc = markdown_parser(skill_doc_file.read_text())
        # get only code blocks
        cls.code_blocks = list(filter(lambda x: x["type"] == "block_code", doc))

    def test_custom_protocol(self):
        """Test the code in the 'Custom protocol' section."""
        # this is the offset of code blocks for the section under testing
        offset = 0
        locals_dict = {}
        compile_and_exec(self.code_blocks[offset]["text"], locals_dict=locals_dict)
        ActualPerformative = locals_dict["Performative"]
        compare_enum_classes(ActualPerformative, DefaultMessage.Performative)

        # load the example of default message of type BYTES
        compile_and_exec(self.code_blocks[offset + 1]["text"], locals_dict=locals_dict)

        # load the definition of the ErrorCode enumeration
        compile_and_exec(self.code_blocks[offset + 2]["text"], locals_dict=locals_dict)
        ExpectedErrorCode = locals_dict["ErrorCode"]
        compare_enum_classes(ExpectedErrorCode, DefaultMessage.ErrorCode)

        # load the example of default message of type ERROR
        _ = compile_and_exec(
            self.code_blocks[offset + 3]["text"], locals_dict=locals_dict
        )

    def test_oef_search_protocol(self):
        """Test the fetchai/oef_search:0.8.0 protocol documentation."""
        # this is the offset of code blocks for the section under testing
        offset = 4

        # define a data model and a description
        locals_dict = {"Enum": Enum}
        compile_and_exec(self.code_blocks[offset]["text"], locals_dict=locals_dict)
        ActualPerformative = locals_dict["Performative"]
        compare_enum_classes(OefSearchMessage.Performative, ActualPerformative)

        compile_and_exec(self.code_blocks[offset + 1]["text"], locals_dict=locals_dict)
        # mind the indexes: +3 before +2
        compile_and_exec(self.code_blocks[offset + 3]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 2]["text"], locals_dict=locals_dict)

        # test the construction of OEF Search Messages does not contain trivial errors.
        locals_dict["OefSearchMessage"] = OefSearchMessage
        compile_and_exec(self.code_blocks[offset + 4]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 5]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 6]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 7]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 8]["text"], locals_dict=locals_dict)
        compile_and_exec(self.code_blocks[offset + 9]["text"], locals_dict=locals_dict)
        # this is just to test that something has actually run
        assert locals_dict["query_data"] == {
            "search_term": "country",
            "search_value": "UK",
            "constraint_type": "==",
        }

        # test the definition of OefErrorOperation
        compile_and_exec(self.code_blocks[offset + 10]["text"], locals_dict=locals_dict)
        ActualOefErrorOperation = locals_dict["OefErrorOperation"]
        ExpectedOefErrorOperation = OefErrorOperation
        compare_enum_classes(ExpectedOefErrorOperation, ActualOefErrorOperation)

    def test_fipa_protocol(self):
        """Test the fetchai/fipa:0.8.0 documentation."""
        offset = 15
        locals_dict = {"Enum": Enum}
        compile_and_exec(self.code_blocks[offset]["text"], locals_dict=locals_dict)
        ActualFipaPerformative = locals_dict["Performative"]
        ExpectedFipaPerformative = FipaMessage.Performative
        compare_enum_classes(ExpectedFipaPerformative, ActualFipaPerformative)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the generic buyer skill."""

from pathlib import Path

from aea.configurations.constants import DEFAULT_LEDGER
from aea.helpers.search.models import Constraint, ConstraintType, Description, Query
from aea.helpers.transaction.base import Terms
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_NAME

from packages.fetchai.skills.generic_buyer.strategy import (
    GenericStrategy,
    SIMPLE_SERVICE_MODEL,
)

from tests.conftest import ROOT_DIR


class TestGenericStrategy(BaseSkillTestCase):
    """Test GenericStrategy of generic buyer."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "generic_buyer")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.ledger_id = DEFAULT_LEDGER
        cls.is_ledger_tx = True
        cls.currency_id = "some_currency_id"
        cls.max_unit_price = 20
        cls.max_tx_fee = 1
        cls.service_id = "some_service_id"
        cls.search_query = {
            "constraint_type": "==",
            "search_key": "seller_service",
            "search_value": "some_search_value",
        }
        cls.location = {
            "longitude": 51.5194,
            "latitude": 0.127,
        }
        cls.search_radius = 5.0
        cls.max_negotiations = 2
        cls.strategy = GenericStrategy(
            ledger_id=cls.ledger_id,
            is_ledger_tx=cls.is_ledger_tx,
            currency_id=cls.currency_id,
            max_unit_price=cls.max_unit_price,
            max_tx_fee=cls.max_tx_fee,
            service_id=cls.service_id,
            search_query=cls.search_query,
            location=cls.location,
            search_radius=cls.search_radius,
            max_negotiations=cls.max_negotiations,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

    def test_properties(self):
        """Test the properties of GenericStrategy class."""
        assert self.strategy.ledger_id == self.ledger_id
        assert self.strategy.is_ledger_tx == self.is_ledger_tx
        assert self.strategy.is_searching is False

        self.strategy.is_searching = True
        assert self.strategy.is_searching is True

        assert self.strategy.balance == 0

        self.strategy.balance = 100
        assert self.strategy.balance == 100

        assert self.strategy.max_negotiations is self.max_negotiations

    def test_get_location_and_service_query(self):
        """Test the get_location_and_service_query method of the GenericStrategy class."""
        query = self.strategy.get_location_and_service_query()

        assert type(query) == Query
        assert len(query.constraints) == 2
        assert query.model is None

        location_constraint = Constraint(
            "location",
            ConstraintType(
                "distance", (self.strategy._agent_location, self.search_radius)
            ),
        )
        assert query.constraints[0] == location_constraint

        service_key_constraint = Constraint(
            self.search_query["search_key"],
            ConstraintType(
                self.search_query["constraint_type"], self.search_query["search_value"],
            ),
        )
        assert query.constraints[1] == service_key_constraint

    def test_get_service_query(self):
        """Test the get_service_query method of the GenericStrategy class."""
        query = self.strategy.get_service_query()

        assert type(query) == Query
        assert len(query.constraints) == 1

        assert query.model == SIMPLE_SERVICE_MODEL

        service_key_constraint = Constraint(
            self.search_query["search_key"],
            ConstraintType(
                self.search_query["constraint_type"], self.search_query["search_value"],
            ),
        )
        assert query.constraints[0] == service_key_constraint

    def test_is_acceptable_proposal(self):
        """Test the is_acceptable_proposal method of the GenericStrategy class."""
        acceptable_description = Description(
            {
                "ledger_id": self.ledger_id,
                "price": 150,
                "currency_id": self.currency_id,
                "service_id": self.service_id,
                "quantity": 10,
                "tx_nonce": "some_tx_nonce",
            }
        )
        is_acceptable = self.strategy.is_acceptable_proposal(acceptable_description)
        assert is_acceptable

        unacceptable_description = Description(
            {
                "ledger_id": self.ledger_id,
                "price": 250,
                "currency_id": self.currency_id,
                "service_id": self.service_id,
                "quantity": 10,
                "tx_nonce": "some_tx_nonce",
            }
        )
        is_acceptable = self.strategy.is_acceptable_proposal(unacceptable_description)
        assert not is_acceptable

    def test_is_affordable_proposal(self):
        """Test the is_affordable_proposal method of the GenericStrategy class."""
        description = Description(
            {
                "ledger_id": self.ledger_id,
                "price": 150,
                "currency_id": self.currency_id,
                "service_id": self.service_id,
                "quantity": 10,
                "tx_nonce": "some_tx_nonce",
            }
        )
        self.strategy.balance = 151
        is_affordable = self.strategy.is_affordable_proposal(description)
        assert is_affordable

        self.strategy.balance = 150
        is_affordable = self.strategy.is_affordable_proposal(description)
        assert not is_affordable

        self.strategy._is_ledger_tx = False
        is_affordable = self.strategy.is_affordable_proposal(description)
        assert is_affordable

    def test_terms_from_proposal(self):
        """Test the terms_from_proposal method of the GenericStrategy class."""
        description = Description(
            {
                "ledger_id": self.ledger_id,
                "price": 150,
                "currency_id": self.currency_id,
                "service_id": self.service_id,
                "quantity": 10,
                "tx_nonce": "some_tx_nonce",
            }
        )
        terms = Terms(
            ledger_id=self.ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address=COUNTERPARTY_NAME,
            amount_by_currency_id={self.currency_id: -150},
            quantities_by_good_id={self.service_id: 10},
            is_sender_payable_tx_fee=True,
            nonce="some_tx_nonce",
            fee_by_currency_id={self.currency_id: self.max_tx_fee},
        )
        assert (
            self.strategy.terms_from_proposal(description, COUNTERPARTY_NAME) == terms
        )
